"""Protocol layer services for XP."""

from xp.models.protocol.conbus_protocol import (
    ConnectionMadeEvent,
    EventTelegramReceivedEvent,
    InvalidTelegramReceivedEvent,
    ModuleDiscoveredEvent,
    TelegramReceivedEvent,
)
from xp.services.protocol.conbus_event_protocol import ConbusEventProtocol

__all__ = ["ConbusEventProtocol"]

# Rebuild models after ConbusEventProtocol is imported to resolve forward references
ConnectionMadeEvent.model_rebuild()
InvalidTelegramReceivedEvent.model_rebuild()
ModuleDiscoveredEvent.model_rebuild()
TelegramReceivedEvent.model_rebuild()
EventTelegramReceivedEvent.model_rebuild()



"""Time parsing utilities for console bus logs."""

import re
from datetime import datetime, time
from typing import Optional


class TimeParsingError(Exception):
    """Raised when time parsing fails."""

    pass


def parse_log_timestamp(
    timestamp_str: str, base_date: Optional[datetime] = None
) -> datetime:
    """Parse timestamp from console bus log format: HH:MM:SS,mmm.

    Args:
        timestamp_str: Timestamp string (e.g., "22:44:20,352")
        base_date: Base date to use (defaults to today)

    Returns:
        datetime object with parsed time

    Raises:
        TimeParsingError: If timestamp format is invalid
    """
    # Pattern: HH:MM:SS,mmm
    pattern = r"^(\d{2}):(\d{2}):(\d{2}),(\d{3})$"
    match = re.match(pattern, timestamp_str.strip())

    if not match:
        raise TimeParsingError(f"Invalid timestamp format: {timestamp_str}")

    try:
        hour = int(match.group(1))
        minute = int(match.group(2))
        second = int(match.group(3))
        millisecond = int(match.group(4))

        # Validate ranges
        if not (0 <= hour <= 23):
            raise TimeParsingError(f"Invalid hour: {hour}")
        if not (0 <= minute <= 59):
            raise TimeParsingError(f"Invalid minute: {minute}")
        if not (0 <= second <= 59):
            raise TimeParsingError(f"Invalid second: {second}")
        if not (0 <= millisecond <= 999):
            raise TimeParsingError(f"Invalid millisecond: {millisecond}")

        # Create time object
        time_obj = time(hour, minute, second, millisecond * 1000)  # microseconds

        # Use base date or today
        if base_date is None:
            date_part = datetime.now().date()
        else:
            date_part = base_date.date()

        return datetime.combine(date_part, time_obj)

    except ValueError as e:
        raise TimeParsingError(f"Error parsing timestamp {timestamp_str}: {e}")


def format_log_timestamp(dt: datetime) -> str:
    """Format datetime to console bus log timestamp format: HH:MM:SS,mmm.

    Args:
        dt: datetime object to format

    Returns:
        Formatted timestamp string
    """
    return dt.strftime("%H:%M:%S,%f")[:-3]  # Remove last 3 digits of microseconds


def parse_time_range(
    time_range_str: str, base_date: Optional[datetime] = None
) -> tuple[datetime, datetime]:
    """
    Parse time range string like "22:44:20,352-22:44:25,500".

    Args:
        time_range_str: Time range string
        base_date: Base date to use

    Returns:
        Tuple of (start_time, end_time)

    Raises:
        TimeParsingError: If format is invalid
    """
    parts = time_range_str.split("-")
    if len(parts) != 2:
        raise TimeParsingError(f"Invalid time range format: {time_range_str}")

    start_time = parse_log_timestamp(parts[0].strip(), base_date)
    end_time = parse_log_timestamp(parts[1].strip(), base_date)

    if start_time > end_time:
        raise TimeParsingError(f"Start time {parts[0]} is after end time {parts[1]}")

    return start_time, end_time


def calculate_duration_ms(start_time: datetime, end_time: datetime) -> int:
    """
    Calculate duration between two timestamps in milliseconds.

    Args:
        start_time: Start timestamp
        end_time: End timestamp

    Returns:
        Duration in milliseconds
    """
    duration = end_time - start_time
    return int(duration.total_seconds() * 1000)


def is_valid_log_timestamp(timestamp_str: str) -> bool:
    """
    Check if timestamp string is valid console bus log format.

    Args:
        timestamp_str: Timestamp string to validate

    Returns:
        True if valid format, False otherwise
    """
    try:
        parse_log_timestamp(timestamp_str)
        return True
    except TimeParsingError:
        return False



"""Unit tests for ConbusOutputService."""

from unittest.mock import MagicMock

import pytest

from xp.models.protocol.conbus_protocol import TelegramReceivedEvent
from xp.models.telegram.action_type import ActionType
from xp.models.telegram.output_telegram import OutputTelegram
from xp.models.telegram.system_function import SystemFunction
from xp.models.telegram.telegram_type import TelegramType
from xp.services.conbus.conbus_output_service import ConbusOutputService
from xp.services.protocol.conbus_event_protocol import ConbusEventProtocol
from xp.services.telegram.telegram_output_service import TelegramOutputService


@pytest.fixture
def mock_conbus_protocol():
    """Create a mock ConbusEventProtocol."""
    protocol = MagicMock(spec=ConbusEventProtocol)
    protocol.timeout_seconds = 5.0
    return protocol


@pytest.fixture
def mock_telegram_output_service():
    """Create a mock TelegramOutputService."""
    service = MagicMock(spec=TelegramOutputService)
    return service


@pytest.fixture
def conbus_output_service(mock_conbus_protocol, mock_telegram_output_service):
    """Create a ConbusOutputService instance for testing."""
    return ConbusOutputService(
        conbus_protocol=mock_conbus_protocol,
        telegram_output_service=mock_telegram_output_service,
    )


def test_init_connects_signals(mock_conbus_protocol, mock_telegram_output_service):
    """Test that __init__ connects all protocol signals."""
    _ = ConbusOutputService(
        conbus_protocol=mock_conbus_protocol,
        telegram_output_service=mock_telegram_output_service,
    )

    # Verify all signals are connected
    mock_conbus_protocol.on_connection_made.connect.assert_called_once()
    mock_conbus_protocol.on_telegram_sent.connect.assert_called_once()
    mock_conbus_protocol.on_telegram_received.connect.assert_called_once()
    mock_conbus_protocol.on_timeout.connect.assert_called_once()
    mock_conbus_protocol.on_failed.connect.assert_called_once()


def test_connection_made_sends_telegram(conbus_output_service, mock_conbus_protocol):
    """Test that connection_made sends action telegram."""
    conbus_output_service.serial_number = "0012345678"
    conbus_output_service.output_number = 5
    conbus_output_service.action_type = ActionType.ON_RELEASE

    conbus_output_service.connection_made()

    mock_conbus_protocol.send_telegram.assert_called_once_with(
        telegram_type=TelegramType.SYSTEM,
        serial_number="0012345678",
        system_function=SystemFunction.ACTION,
        data_value="05AB",
    )


def test_telegram_sent_updates_response(conbus_output_service):
    """Test that telegram_sent updates service response."""
    telegram = "test_telegram"

    conbus_output_service.telegram_sent(telegram)

    assert conbus_output_service.service_response.sent_telegram == telegram


def test_telegram_received_appends_to_list(conbus_output_service, mock_conbus_protocol):
    """Test that telegram_received appends to received_telegrams list."""
    event = TelegramReceivedEvent(
        protocol=mock_conbus_protocol,
        frame="test_frame",
        telegram="test_telegram",
        telegram_type=TelegramType.REPLY.value,
        serial_number="0012345678",
        payload="test_payload",
        checksum="00",
        checksum_valid=True,
    )

    conbus_output_service.telegram_received(event)

    assert conbus_output_service.service_response.received_telegrams == ["test_frame"]


def test_telegram_received_processes_ack(
    conbus_output_service, mock_telegram_output_service, mock_conbus_protocol
):
    """Test that telegram_received processes ACK response."""
    conbus_output_service.serial_number = "0012345678"
    output_telegram = OutputTelegram(
        serial_number="0012345678",
        system_function=SystemFunction.ACK,
        checksum="00",
        raw_telegram="<S0012345678F16D>",
    )
    mock_telegram_output_service.parse_reply_telegram.return_value = output_telegram

    event = TelegramReceivedEvent(
        protocol=mock_conbus_protocol,
        frame="test_frame",
        telegram="test_telegram",
        telegram_type=TelegramType.REPLY.value,
        serial_number="0012345678",
        payload="test_payload",
        checksum="00",
        checksum_valid=True,
    )

    # Track signal emission
    signal_emitted: list = []
    conbus_output_service.on_finish.connect(signal_emitted.append)

    conbus_output_service.telegram_received(event)

    # Verify ACK was processed
    assert len(signal_emitted) == 1
    assert signal_emitted[0].success is True


def test_timeout_calls_failed(conbus_output_service):
    """Test that timeout calls failed with timeout message."""
    # Track signal emission
    signal_emitted: list = []
    conbus_output_service.on_finish.connect(signal_emitted.append)

    conbus_output_service.timeout()

    # Verify failed was called with timeout message
    assert len(signal_emitted) == 1
    assert signal_emitted[0].success is False
    assert signal_emitted[0].error == "Timeout"


def test_failed_emits_signal(conbus_output_service):
    """Test that failed emits on_finish signal."""
    # Track signal emission
    signal_emitted: list = []
    conbus_output_service.on_finish.connect(signal_emitted.append)

    conbus_output_service.failed("Test error")

    # Verify signal was emitted with error
    assert len(signal_emitted) == 1
    assert signal_emitted[0].success is False
    assert signal_emitted[0].error == "Test error"


def test_send_action_sets_state(conbus_output_service, mock_conbus_protocol):
    """Test that send_action sets service state."""
    conbus_output_service.send_action(
        serial_number="0012345678",
        output_number=10,
        action_type=ActionType.OFF_PRESS,
        timeout_seconds=3.0,
    )

    assert conbus_output_service.serial_number == "0012345678"
    assert conbus_output_service.output_number == 10
    assert conbus_output_service.action_type == ActionType.OFF_PRESS
    assert mock_conbus_protocol.timeout_seconds == 3.0


def test_set_timeout_delegates_to_protocol(conbus_output_service, mock_conbus_protocol):
    """Test that set_timeout delegates to protocol."""
    conbus_output_service.set_timeout(10.0)

    assert mock_conbus_protocol.timeout_seconds == 10.0


def test_start_reactor_delegates_to_protocol(
    conbus_output_service, mock_conbus_protocol
):
    """Test that start_reactor delegates to protocol."""
    conbus_output_service.start_reactor()

    mock_conbus_protocol.start_reactor.assert_called_once()


def test_stop_reactor_delegates_to_protocol(
    conbus_output_service, mock_conbus_protocol
):
    """Test that stop_reactor delegates to protocol."""
    conbus_output_service.stop_reactor()

    mock_conbus_protocol.stop_reactor.assert_called_once()


def test_enter_resets_state(conbus_output_service):
    """Test that __enter__ resets state for singleton reuse."""
    # Set some state
    conbus_output_service.service_response.success = True
    conbus_output_service.output_state = "test"

    # Enter context
    result = conbus_output_service.__enter__()

    # Verify state reset
    assert result is conbus_output_service
    assert conbus_output_service.service_response.success is False
    assert conbus_output_service.output_state == ""


def test_exit_disconnects_signals(conbus_output_service, mock_conbus_protocol):
    """Test that __exit__ disconnects all signals and stops reactor."""
    conbus_output_service.__exit__(None, None, None)

    # Verify protocol signals disconnected
    mock_conbus_protocol.on_connection_made.disconnect.assert_called_once()
    mock_conbus_protocol.on_telegram_sent.disconnect.assert_called_once()
    mock_conbus_protocol.on_telegram_received.disconnect.assert_called_once()
    mock_conbus_protocol.on_timeout.disconnect.assert_called_once()
    mock_conbus_protocol.on_failed.disconnect.assert_called_once()

    # Verify reactor stopped
    mock_conbus_protocol.stop_reactor.assert_called_once()


def test_succeed_emits_signal_with_output_telegram(conbus_output_service):
    """Test that succeed emits signal with successful response."""
    conbus_output_service.serial_number = "0012345678"
    conbus_output_service.output_number = 7
    conbus_output_service.action_type = ActionType.ON_RELEASE

    output_telegram = OutputTelegram(
        serial_number="0012345678",
        system_function=SystemFunction.ACK,
        checksum="00",
        raw_telegram="<S0012345678F16D>",
    )

    # Track signal emission
    signal_emitted: list = []
    conbus_output_service.on_finish.connect(signal_emitted.append)

    conbus_output_service.succeed(output_telegram)

    # Verify signal was emitted with success
    assert len(signal_emitted) == 1
    assert signal_emitted[0].success is True
    assert signal_emitted[0].serial_number == "0012345678"
    assert signal_emitted[0].output_number == 7
    assert signal_emitted[0].action_type == ActionType.ON_RELEASE
    assert signal_emitted[0].output_telegram == output_telegram



"""Conbus event operations CLI commands."""

import json

import click

from xp.cli.commands.conbus.conbus import conbus, conbus_event
from xp.cli.utils.decorators import connection_command
from xp.cli.utils.module_type_choice import MODULE_TYPE
from xp.models import ConbusEventRawResponse
from xp.services.conbus.conbus_event_list_service import ConbusEventListService
from xp.services.conbus.conbus_event_raw_service import ConbusEventRawService


@conbus_event.command("list")
@click.pass_context
def list_events(ctx: click.Context) -> None:
    r"""
    List configured event telegrams from module action tables.

    Reads conson.yml configuration, parses action tables, and groups
    modules by their event keys to show which modules are assigned to
    each event (button configuration).

    Output is sorted by module count (most frequently used events first).

    Args:
        ctx: Click context object.

    Examples:
        \b
        xp conbus event list
    """
    service: ConbusEventListService = (
        ctx.obj.get("container").get_container().resolve(ConbusEventListService)
    )
    click.echo(json.dumps(service.list_events().to_dict(), indent=2))


@conbus_event.command("raw")
@click.argument("module_type", type=MODULE_TYPE)
@click.argument("link_number", type=click.IntRange(0, 99))
@click.argument("input_number", type=click.IntRange(0, 9))
@click.argument("time_ms", type=click.IntRange(min=1), default=1000)
@click.pass_context
@connection_command()
def send_event_raw(
    ctx: click.Context,
    module_type: int,
    link_number: int,
    input_number: int,
    time_ms: int,
) -> None:
    r"""
    Send raw event telegrams to simulate button presses.

    Args:
        ctx: Click context object.
        module_type: Module type code (e.g., CP20, XP33).
        link_number: Link number (0-99).
        input_number: Input number (0-9).
        time_ms: Delay between MAKE/BREAK events in milliseconds (default: 1000).

    Examples:
        \b
        xp conbus event raw CP20 00 00
        xp conbus event raw XP33 00 00 500
    """

    def on_finish(response: ConbusEventRawResponse) -> None:
        """
        Handle successful completion of event raw operation.

        Args:
            response: Event raw response with sent and received telegrams.
        """
        click.echo(json.dumps(response.to_dict(), indent=2))

    def on_progress(telegram: str) -> None:
        """
        Handle progress updates during event operation.

        Args:
            telegram: Received telegram.
        """
        click.echo(json.dumps({"telegram": telegram}))

    service: ConbusEventRawService = (
        ctx.obj.get("container").get_container().resolve(ConbusEventRawService)
    )
    service.run(
        module_type_code=module_type,
        link_number=link_number,
        input_number=input_number,
        time_ms=time_ms,
        progress_callback=on_progress,
        finish_callback=on_finish,
        timeout_seconds=5,
    )
    service.start_reactor()


# Register the event command group with conbus
conbus.add_command(conbus_event)



"""Common decorators for CLI commands."""

import functools
from typing import Any, Callable, Tuple, Type

import click

from xp.cli.utils.formatters import OutputFormatter


def handle_service_errors(
    *service_exceptions: Type[Exception],
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Handle common service exceptions with consistent JSON error formatting.

    Args:
        service_exceptions: Tuple of exception types to catch and handle.

    Returns:
        Decorator function that wraps commands with error handling.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Apply error handling to the decorated function.

        Args:
            func: The function to decorate.

        Returns:
            Wrapped function with error handling.
        """

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """
            Execute function with error handling.

            Args:
                args: Positional arguments passed to the decorated function.
                kwargs: Keyword arguments passed to the decorated function.

            Returns:
                Result from the decorated function.

            Raises:
                SystemExit: When a service exception or unexpected error occurs.
            """
            formatter = OutputFormatter(True)

            try:
                return func(*args, **kwargs)
            except service_exceptions as e:
                error_response = formatter.error_response(str(e))
                click.echo(error_response)
                raise SystemExit(1)
            except Exception as e:
                # Handle unexpected errors
                error_response = formatter.error_response(f"Unexpected error: {e}")
                click.echo(error_response)
                raise SystemExit(1)

        return wrapper

    return decorator


def common_options(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Add validation option to command.

    Args:
        func: The function to decorate.

    Returns:
        Decorated function with common options.
    """
    return func


def telegram_parser_command(
    service_exceptions: Tuple[Type[Exception], ...] = (),
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Apply telegram parsing commands with standard error handling.

    Args:
        service_exceptions: Additional service exceptions to handle.

    Returns:
        Decorator function for telegram parsing commands.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Apply telegram parser decorators.

        Args:
            func: The function to decorate.

        Returns:
            Decorated function with telegram parsing support.
        """
        # Apply common options
        func = common_options(func)

        # Apply error handling for telegram parsing
        from xp.services.telegram.telegram_service import TelegramParsingError

        exceptions = (TelegramParsingError,) + service_exceptions
        func = handle_service_errors(*exceptions)(func)

        return func

    return decorator


def service_command(
    *service_exceptions: Type[Exception],
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Apply service-based commands with error handling and JSON output.

    Args:
        service_exceptions: Service exception types to handle.

    Returns:
        Decorator function for service commands.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Apply service command decorators.

        Args:
            func: The function to decorate.

        Returns:
            Decorated function with service error handling.
        """
        func = handle_service_errors(*service_exceptions)(func)
        return func

    return decorator


def list_command(
    *service_exceptions: Type[Exception],
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Apply list/search commands with common options.

    Args:
        service_exceptions: Service exception types to handle.

    Returns:
        Decorator function for list commands.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Apply list command decorators.

        Args:
            func: The function to decorate.

        Returns:
            Decorated function with list error handling.
        """
        func = handle_service_errors(*service_exceptions)(func)
        return func

    return decorator


def file_operation_command() -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Apply file operation commands with common filters.

    Returns:
        Decorator function for file operation commands.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Apply file operation decorators.

        Args:
            func: The function to decorate.

        Returns:
            Decorated function with filter options.
        """
        func = click.option(
            "--time-range", help="Filter by time range (HH:MM:SS,mmm-HH:MM:SS,mmm)"
        )(func)
        func = click.option(
            "--filter-direction",
            type=click.Choice(["tx", "rx"]),
            help="Filter by direction",
        )(func)
        func = click.option(
            "--filter-type",
            type=click.Choice(["event", "system", "reply"]),
            help="Filter by telegram type",
        )(func)
        return func

    return decorator


def with_formatter(
    formatter_class: Any = None,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Inject a formatter instance into the command.

    Args:
        formatter_class: Custom formatter class to use.

    Returns:
        Decorator function that injects a formatter.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Apply formatter injection.

        Args:
            func: The function to decorate.

        Returns:
            Wrapped function with formatter injection.
        """

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """
            Execute function with injected formatter.

            Args:
                args: Positional arguments passed to the decorated function.
                kwargs: Keyword arguments passed to the decorated function.

            Returns:
                Result from the decorated function.
            """
            formatter_cls = formatter_class or OutputFormatter
            formatter = formatter_cls(True)
            kwargs["formatter"] = formatter
            return func(*args, **kwargs)

        return wrapper

    return decorator


def require_arguments(
    *required_args: str,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Validate required arguments are present.

    Args:
        required_args: Names of required arguments.

    Returns:
        Decorator function that validates required arguments.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Apply argument validation.

        Args:
            func: The function to decorate.

        Returns:
            Wrapped function with argument validation.
        """

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """
            Execute function with argument validation.

            Args:
                args: Positional arguments passed to the decorated function.
                kwargs: Keyword arguments passed to the decorated function.

            Returns:
                Result from the decorated function.

            Raises:
                SystemExit: When required arguments are missing.
            """
            formatter = OutputFormatter(True)

            # Check for missing required arguments
            missing_args = [
                arg_name
                for arg_name in required_args
                if arg_name in kwargs and kwargs[arg_name] is None
            ]

            if missing_args:
                error_msg = f"Missing required arguments: {', '.join(missing_args)}"
                error_response = formatter.error_response(error_msg)
                click.echo(error_response)
                raise SystemExit(1)

            return func(*args, **kwargs)

        return wrapper

    return decorator


def connection_command() -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Apply commands that connect to remote services.

    Returns:
        Decorator function for connection commands.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Apply connection command decorators.

        Args:
            func: The function to decorate.

        Returns:
            Wrapped function with connection error handling.
        """

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """
            Execute function with connection error handling.

            Args:
                args: Positional arguments passed to the decorated function.
                kwargs: Keyword arguments passed to the decorated function.

            Returns:
                Result from the decorated function.

            Raises:
                SystemExit: When a connection timeout occurs.
                Exception: Re-raises other exceptions for handling by other decorators.
            """
            formatter = OutputFormatter(True)

            try:
                return func(*args, **kwargs)
            except Exception as e:
                if "Connection timeout" in str(e):
                    # Special handling for connection timeouts
                    error_msg = "Connection timeout - server may be unreachable"
                    error_response = formatter.error_response(error_msg)
                    click.echo(error_response)
                    raise SystemExit(1)
                else:
                    # Re-raise other exceptions to be handled by other decorators
                    raise

        return wrapper

    return decorator



"""Unit tests for XP24 Action Table short format."""

import pytest

from xp.models.actiontable.msactiontable_xp24 import InputAction, Xp24MsActionTable
from xp.models.telegram.input_action_type import InputActionType
from xp.models.telegram.timeparam_type import TimeParam


class TestXp24ShortFormat:
    """Test cases for XP24 action table short format conversion."""

    def test_to_short_format_basic(self):
        """Test basic conversion to short format."""
        action_table = Xp24MsActionTable(
            input1_action=InputAction(
                type=InputActionType.TOGGLE, param=TimeParam.T05SEC
            ),
            input2_action=InputAction(
                type=InputActionType.TOGGLE, param=TimeParam.T1SEC
            ),
            input3_action=InputAction(
                type=InputActionType.TOGGLE, param=TimeParam.NONE
            ),
            input4_action=InputAction(
                type=InputActionType.TOGGLE, param=TimeParam.NONE
            ),
        )

        assert action_table.to_short_format() == [
            "T:1 T:2 T:0 T:0 | M12:0 M34:0 C12:0 C34:0 DT:12"
        ]

    def test_to_short_format_with_settings(self):
        """Test conversion to short format with settings."""
        short = Xp24MsActionTable(
            input1_action=InputAction(
                type=InputActionType.TOGGLE, param=TimeParam.NONE
            ),
            input2_action=InputAction(
                type=InputActionType.TOGGLE, param=TimeParam.NONE
            ),
            input3_action=InputAction(
                type=InputActionType.TOGGLE, param=TimeParam.NONE
            ),
            input4_action=InputAction(
                type=InputActionType.TOGGLE, param=TimeParam.NONE
            ),
            mutex12=True,
            mutex34=True,
            curtain12=False,
            curtain34=False,
            mutual_deadtime=20,
        ).to_short_format()
        assert short == ["T:0 T:0 T:0 T:0 | M12:1 M34:1 C12:0 C34:0 DT:20"]

    def test_to_short_format_mixed_actions(self):
        """Test conversion with mixed action types."""
        short = Xp24MsActionTable(
            input1_action=InputAction(type=InputActionType.ON, param=TimeParam.T5SEC),
            input2_action=InputAction(type=InputActionType.OFF, param=TimeParam.NONE),
            input3_action=InputAction(
                type=InputActionType.LEVELSET, param=TimeParam.T5MIN
            ),
            input4_action=InputAction(
                type=InputActionType.SCENESET, param=TimeParam.T2MIN
            ),
        ).to_short_format()
        assert short == ["ON:4 OF:0 LS:12 SS:11 | M12:0 M34:0 C12:0 C34:0 DT:12"]

    def test_to_short_format_all_void(self):
        """Test conversion with all VOID actions."""
        short = Xp24MsActionTable(
            input1_action=InputAction(type=InputActionType.VOID, param=TimeParam.NONE),
            input2_action=InputAction(type=InputActionType.VOID, param=TimeParam.NONE),
            input3_action=InputAction(type=InputActionType.VOID, param=TimeParam.NONE),
            input4_action=InputAction(type=InputActionType.VOID, param=TimeParam.NONE),
        ).to_short_format()
        assert short == ["V:0 V:0 V:0 V:0 | M12:0 M34:0 C12:0 C34:0 DT:12"]

    def test_from_short_format_basic(self):
        """Test parsing basic short format."""
        short = "T:1 T:2 T:0 T:0 | M12:0 M34:0 C12:0 C34:0 DT:12"
        action_table = Xp24MsActionTable.from_short_format([short])

        assert action_table.input1_action.type == InputActionType.TOGGLE
        assert action_table.input1_action.param == TimeParam.T05SEC
        assert action_table.input2_action.type == InputActionType.TOGGLE
        assert action_table.input2_action.param == TimeParam.T1SEC
        assert action_table.input3_action.type == InputActionType.TOGGLE
        assert action_table.input3_action.param == TimeParam.NONE
        assert action_table.input4_action.type == InputActionType.TOGGLE
        assert action_table.input4_action.param == TimeParam.NONE

    def test_from_short_format_with_settings(self):
        """Test parsing short format with settings."""
        short = "T:0 T:0 T:0 T:0 | M12:1 M34:1 C12:0 C34:0 DT:20"
        action_table = Xp24MsActionTable.from_short_format([short])

        assert action_table.mutex12 is True
        assert action_table.mutex34 is True
        assert action_table.curtain12 is False
        assert action_table.curtain34 is False
        assert action_table.mutual_deadtime == 20

    def test_from_short_format_mixed_actions(self):
        """Test parsing mixed action types."""
        short = "ON:4 OF:0 LS:12 SS:11 | M12:0 M34:0 C12:0 C34:0 DT:12"
        action_table = Xp24MsActionTable.from_short_format([short])

        assert action_table.input1_action.type == InputActionType.ON
        assert action_table.input1_action.param == TimeParam.T5SEC
        assert action_table.input2_action.type == InputActionType.OFF
        assert action_table.input2_action.param == TimeParam.NONE
        assert action_table.input3_action.type == InputActionType.LEVELSET
        assert action_table.input3_action.param == TimeParam.T5MIN
        assert action_table.input4_action.type == InputActionType.SCENESET
        assert action_table.input4_action.param == TimeParam.T2MIN

    def test_round_trip_conversion(self):
        """Test that converting to short and back preserves data."""
        original = Xp24MsActionTable(
            input1_action=InputAction(
                type=InputActionType.TOGGLE, param=TimeParam.T05SEC
            ),
            input2_action=InputAction(type=InputActionType.ON, param=TimeParam.T5SEC),
            input3_action=InputAction(
                type=InputActionType.LEVELSET, param=TimeParam.T5MIN
            ),
            input4_action=InputAction(
                type=InputActionType.SCENESET, param=TimeParam.T2MIN
            ),
            mutex12=True,
            mutex34=False,
            curtain12=True,
            curtain34=False,
            mutual_deadtime=20,
        )

        short = original.to_short_format()
        restored = Xp24MsActionTable.from_short_format(short)

        assert restored == original

    def test_from_short_format_invalid_format(self):
        """Test that invalid format raises ValueError."""
        with pytest.raises(ValueError, match="Invalid short format"):
            Xp24MsActionTable.from_short_format(
                ["INVALID | M12:0 M34:0 C12:0 C34:0 DT:12"]
            )

    def test_from_short_format_invalid_action_code(self):
        """Test that invalid action code raises ValueError."""
        with pytest.raises(ValueError, match="Unknown action code"):
            Xp24MsActionTable.from_short_format(
                ["XX:0 T:0 T:0 T:0 | M12:0 M34:0 C12:0 C34:0 DT:12"]
            )

    def test_from_short_format_invalid_param(self):
        """Test that invalid param raises ValueError."""
        with pytest.raises(ValueError, match="Invalid time param"):
            Xp24MsActionTable.from_short_format(
                ["T:999 T:0 T:0 T:0 | M12:0 M34:0 C12:0 C34:0 DT:12"]
            )

    def test_from_short_format_missing_colon(self):
        """Test that missing colon raises ValueError."""
        with pytest.raises(ValueError, match="Invalid action format"):
            Xp24MsActionTable.from_short_format(
                ["T1 T:0 T:0 T:0 | M12:0 M34:0 C12:0 C34:0 DT:12"]
            )

    def test_all_action_types(self):
        """Test that all action types can be converted."""
        action_types = [
            (InputActionType.VOID, "V"),
            (InputActionType.ON, "ON"),
            (InputActionType.OFF, "OF"),
            (InputActionType.TOGGLE, "T"),
            (InputActionType.BLOCK, "BL"),
            (InputActionType.AUXRELAY, "AX"),
            (InputActionType.MUTUALEX, "MX"),
            (InputActionType.LEVELUP, "LU"),
            (InputActionType.LEVELDOWN, "LD"),
            (InputActionType.LEVELINC, "LI"),
            (InputActionType.LEVELDEC, "LC"),
            (InputActionType.LEVELSET, "LS"),
            (InputActionType.FADETIME, "FT"),
            (InputActionType.SCENESET, "SS"),
            (InputActionType.SCENENEXT, "SN"),
            (InputActionType.SCENEPREV, "SP"),
            (InputActionType.CTRLMETHOD, "CM"),
            (InputActionType.RETURNDATA, "RD"),
            (InputActionType.DELAYEDON, "DO"),
            (InputActionType.EVENTTIMER1, "E1"),
            (InputActionType.EVENTTIMER2, "E2"),
            (InputActionType.EVENTTIMER3, "E3"),
            (InputActionType.EVENTTIMER4, "E4"),
            (InputActionType.STEPCTRL, "SC"),
            (InputActionType.STEPCTRLUP, "SU"),
            (InputActionType.STEPCTRLDOWN, "SD"),
            (InputActionType.LEVELSETINTERN, "LN"),
            (InputActionType.FADE, "FD"),
            (InputActionType.LEARN, "LR"),
        ]

        for action_type, expected_code in action_types:
            action_table = Xp24MsActionTable(
                input1_action=InputAction(type=action_type, param=TimeParam.NONE),
            )
            short = action_table.to_short_format()
            assert short[0].startswith(f"{expected_code}:0 ")

            # Test round-trip
            restored = Xp24MsActionTable.from_short_format(short)
            assert restored.input1_action.type == action_type

    def test_all_time_params(self):
        """Test that all time param values can be converted."""
        time_params = [
            TimeParam.NONE,
            TimeParam.T05SEC,
            TimeParam.T1SEC,
            TimeParam.T2SEC,
            TimeParam.T5SEC,
            TimeParam.T10SEC,
            TimeParam.T15SEC,
            TimeParam.T20SEC,
            TimeParam.T30SEC,
            TimeParam.T45SEC,
            TimeParam.T1MIN,
            TimeParam.T2MIN,
            TimeParam.T5MIN,
            TimeParam.T10MIN,
            TimeParam.T15MIN,
            TimeParam.T20MIN,
            TimeParam.T30MIN,
            TimeParam.T45MIN,
            TimeParam.T60MIN,
            TimeParam.T120MIN,
        ]

        for time_param in time_params:
            action_table = Xp24MsActionTable(
                input1_action=InputAction(
                    type=InputActionType.TOGGLE, param=time_param
                ),
            )
            short = action_table.to_short_format()
            assert short == (
                [f"T:{time_param.value} T:0 T:0 T:0 | M12:0 M34:0 C12:0 C34:0 DT:12"]
            )

            # Test round-trip
            restored = Xp24MsActionTable.from_short_format(short)
            assert restored.input1_action.param == time_param



"""Conbus export service for exporting device configurations."""

import asyncio
import logging
from pathlib import Path
from typing import Any, Optional

import yaml
from psygnal import Signal

from xp.models.conbus.conbus_export import ConbusExportResponse
from xp.models.config.conson_module_config import (
    ConsonModuleConfig,
    ConsonModuleListConfig,
)
from xp.models.protocol.conbus_protocol import TelegramReceivedEvent
from xp.models.telegram.datapoint_type import DataPointType
from xp.models.telegram.reply_telegram import ReplyTelegram
from xp.models.telegram.system_function import SystemFunction
from xp.models.telegram.telegram_type import TelegramType
from xp.services.protocol.conbus_event_protocol import ConbusEventProtocol
from xp.services.telegram.telegram_service import TelegramService


class ConbusExportService:
    """
    Service for exporting Conbus device configurations.

    Discovers all devices on the Conbus network and queries their configuration
    datapoints to generate a structured export file compatible with conson.yml format.

    Attributes:
        conbus_protocol: Protocol for Conbus communication.
        discovered_devices: List of discovered device serial numbers.
        device_configs: Device configurations (ConsonModuleConfig instances).
        export_result: Final export result.
        export_status: Export status (OK, FAILED_TIMEOUT, etc.).
        on_progress: Signal emitted on device discovery (serial, current, total).
        on_device_exported: Signal emitted when device export completes.
        on_finish: Signal emitted when export finishes.
        DATAPOINT_SEQUENCE: Sequence of 7 datapoints to query for each device.
    """

    # Signals (class attributes)
    on_progress: Signal = Signal(str, int, int)  # serial, current, total
    on_device_exported: Signal = Signal(ConsonModuleConfig)
    on_finish: Signal = Signal(ConbusExportResponse)

    # Datapoint sequence to query for each device
    DATAPOINT_SEQUENCE = [
        DataPointType.MODULE_TYPE,
        DataPointType.MODULE_TYPE_CODE,
        DataPointType.LINK_NUMBER,
        DataPointType.MODULE_NUMBER,
        DataPointType.SW_VERSION,
        DataPointType.HW_VERSION,
        DataPointType.AUTO_REPORT_STATUS,
    ]

    def __init__(
        self, conbus_protocol: ConbusEventProtocol, telegram_service: TelegramService
    ) -> None:
        """
        Initialize the Conbus export service.

        Args:
            conbus_protocol: Protocol for Conbus communication.
            telegram_service: TelegramService for telegram parsing.
        """
        self.logger = logging.getLogger(__name__)
        self.conbus_protocol = conbus_protocol
        self.telegram_service = telegram_service

        # State management
        self.discovered_devices: list[str] = []
        self.device_configs: dict[str, ConsonModuleConfig] = {}
        self.export_result = ConbusExportResponse(success=False)
        self.export_status = "OK"
        self._finalized = False  # Track if export has been finalized

        # Connect protocol signals
        self.conbus_protocol.on_connection_made.connect(self.connection_made)
        self.conbus_protocol.on_telegram_sent.connect(self.telegram_sent)
        self.conbus_protocol.on_telegram_received.connect(self.telegram_received)
        self.conbus_protocol.on_timeout.connect(self.timeout)
        self.conbus_protocol.on_failed.connect(self.failed)

    def connection_made(self) -> None:
        """Handle connection established event."""
        self.logger.debug("Connection established, starting discovery")

        # Send DISCOVERY telegram
        self.conbus_protocol.send_telegram(
            telegram_type=TelegramType.SYSTEM,
            serial_number="0000000000",
            system_function=SystemFunction.DISCOVERY,
            data_value="00",
        )

    def telegram_sent(self, telegram: str) -> None:
        """
        Handle telegram sent event.

        Args:
            telegram: Telegram that was sent.
        """
        self.export_result.sent_telegrams.append(telegram)

    def telegram_received(self, event: TelegramReceivedEvent) -> None:
        """
        Handle telegram received event.

        Args:
            event: Telegram received event.
        """
        self.export_result.received_telegrams.append(event.telegram)

        # Only process valid reply telegrams
        if not event.checksum_valid or event.telegram_type != TelegramType.REPLY.value:
            return

        # Parse telegram using TelegramService
        try:
            parsed: ReplyTelegram = self.telegram_service.parse_reply_telegram(
                event.frame
            )
        except Exception as e:
            self.logger.debug(f"Failed to parse telegram: {e}")
            return

        # Check for discovery response (F01D)
        if parsed.system_function == SystemFunction.DISCOVERY:
            self._handle_discovery_response(parsed.serial_number)

        # Check for datapoint response (F02D)
        elif parsed.system_function == SystemFunction.READ_DATAPOINT:
            if parsed.datapoint_type and parsed.data_value:
                self._handle_datapoint_response(
                    parsed.serial_number, parsed.datapoint_type.value, parsed.data_value
                )

    def _handle_discovery_response(self, serial_number: str) -> None:
        """
        Handle discovery response and query all datapoints.

        Args:
            serial_number: Serial number of discovered device.
        """
        if serial_number in self.discovered_devices:
            self.logger.debug(f"Ignoring duplicate discovery: {serial_number}")
            return

        self.logger.debug(f"Device discovered: {serial_number}")
        self.discovered_devices.append(serial_number)

        # Create ConsonModuleConfig with placeholder values for required fields
        module = ConsonModuleConfig(
            name="UNKNOWN",  # Will be updated when link_number arrives
            serial_number=serial_number,
            module_type="UNKNOWN",  # Required field
            module_type_code=0,  # Required field
            link_number=0,  # Required field
        )
        self.device_configs[serial_number] = module

        # Emit progress signal
        current = len(self.discovered_devices)
        total = current  # We don't know total until timeout
        self.on_progress.emit(serial_number, current, total)

        # Send all datapoint queries immediately (protocol handles throttling)
        self.logger.debug(
            f"Sending {len(self.DATAPOINT_SEQUENCE)} queries for {serial_number}"
        )
        for datapoint in self.DATAPOINT_SEQUENCE:
            self.conbus_protocol.send_telegram(
                telegram_type=TelegramType.SYSTEM,
                serial_number=serial_number,
                system_function=SystemFunction.READ_DATAPOINT,
                data_value=datapoint.value,
            )

    def _handle_datapoint_response(
        self, serial_number: str, datapoint_code: str, value: str
    ) -> None:
        """
        Handle datapoint response and store value.

        Args:
            serial_number: Serial number of device.
            datapoint_code: Datapoint type code.
            value: Datapoint value.
        """
        if serial_number not in self.device_configs:
            self.logger.warning(
                f"Received datapoint for unknown device: {serial_number}"
            )
            return

        self.logger.debug(f"Datapoint {datapoint_code}={value} for {serial_number}")

        # Store value in device config
        datapoint = DataPointType.from_code(datapoint_code)
        if datapoint:
            self._store_datapoint_value(serial_number, datapoint, value)
            self._check_device_complete(serial_number)
        else:
            self.logger.warning(f"Unknown datapoint code: {datapoint_code}")

    def _store_datapoint_value(
        self, serial_number: str, datapoint: DataPointType, value: str
    ) -> None:
        """
        Store datapoint value in device config.

        Args:
            serial_number: Serial number of device.
            datapoint: Datapoint type.
            value: Datapoint value.
        """
        module = self.device_configs[serial_number]

        try:
            if datapoint == DataPointType.MODULE_TYPE:
                module.module_type = value
            elif datapoint == DataPointType.MODULE_TYPE_CODE:
                module.module_type_code = int(value)
            elif datapoint == DataPointType.LINK_NUMBER:
                link = int(value)
                module.link_number = link
                module.name = f"A{link}"
            elif datapoint == DataPointType.MODULE_NUMBER:
                module.module_number = int(value)
            elif datapoint == DataPointType.SW_VERSION:
                module.sw_version = value
            elif datapoint == DataPointType.HW_VERSION:
                module.hw_version = value
            elif datapoint == DataPointType.AUTO_REPORT_STATUS:
                module.auto_report_status = value
        except (ValueError, TypeError) as e:
            self.logger.warning(f"Invalid value '{value}' for {datapoint.name}: {e}")

    def _is_device_complete(self, serial_number: str) -> bool:
        """
        Check if a device has all required datapoints.

        Args:
            serial_number: Serial number of device.

        Returns:
            True if device is complete, False otherwise.
        """
        module = self.device_configs[serial_number]
        return all(
            [
                module.module_type not in ("UNKNOWN", None, ""),
                module.module_type_code is not None and module.module_type_code > 0,
                module.link_number is not None and module.link_number > 0,
                module.sw_version is not None,
                module.hw_version is not None,
                module.auto_report_status is not None,
                module.module_number is not None,
            ]
        )

    def _check_device_complete(self, serial_number: str) -> None:
        """
        Check if device has all datapoints and emit completion signal.

        Args:
            serial_number: Serial number of device.
        """
        if self._is_device_complete(serial_number):
            self.logger.debug(f"Device {serial_number} complete (7/7 datapoints)")
            module = self.device_configs[serial_number]
            self.on_device_exported.emit(module)

            # Check if all devices complete
            if all(self._is_device_complete(sn) for sn in self.discovered_devices):
                self.logger.debug("All devices complete")
                self._finalize_export()

    def _finalize_export(self) -> None:
        """Finalize export and write file."""
        # Only finalize once
        if self._finalized:
            return

        self._finalized = True
        self.logger.info("Finalizing export")

        if not self.discovered_devices:
            self.export_status = "FAILED_NO_DEVICES"
            self.export_result.success = False
            self.export_result.error = "No devices found"
            self.export_result.export_status = self.export_status
            self.on_finish.emit(self.export_result)
            return

        # Convert dict values to list (already ConsonModuleConfig instances!)
        modules = list(self.device_configs.values())

        # Sort modules by link_number
        modules.sort(key=lambda m: m.link_number if m.link_number is not None else 999)

        # Create ConsonModuleListConfig
        try:
            module_list = ConsonModuleListConfig(root=modules)
            self.export_result.config = module_list
            self.export_result.device_count = len(modules)

            # Write to file
            self._write_export_file("export.yml")

            self.export_result.success = True
            self.export_result.export_status = self.export_status
            self.on_finish.emit(self.export_result)

        except Exception as e:
            self.logger.error(f"Failed to create export: {e}")
            self.export_status = "FAILED_WRITE"
            self.export_result.success = False
            self.export_result.error = str(e)
            self.export_result.export_status = self.export_status
            self.on_finish.emit(self.export_result)

    def _write_export_file(self, path: str) -> None:
        """
        Write export to YAML file.

        Args:
            path: Output file path.

        Raises:
            Exception: If file write fails.
        """
        try:
            output_path = Path(path)

            if self.export_result.config:
                # Use Pydantic's model_dump to serialize, excluding only internal fields
                data = self.export_result.config.model_dump(
                    exclude={
                        "root": {
                            "__all__": {
                                "enabled",
                                "conbus_ip",
                                "conbus_port",
                                "action_table",
                            }
                        }
                    },
                    exclude_none=True,
                )

                # Export as list at root level (not wrapped in 'root:' key)
                modules_list = data.get("root", [])

                with output_path.open("w") as f:
                    # Dump each module separately with blank lines between them
                    for i, module in enumerate(modules_list):
                        # Add blank line before each module except the first
                        if i > 0:
                            f.write("\n")

                        # Dump single item as list element
                        yaml_str = yaml.safe_dump(
                            [module],
                            default_flow_style=False,
                            sort_keys=False,
                            allow_unicode=True,
                        )
                        # Remove the trailing newline and write
                        f.write(yaml_str.rstrip("\n") + "\n")

            self.logger.info(f"Export written to {path}")
            self.export_result.output_file = path

        except Exception as e:
            self.logger.error(f"Failed to write export file: {e}")
            self.export_status = "FAILED_WRITE"
            raise

    def timeout(self) -> None:
        """Handle timeout event."""
        timeout = self.conbus_protocol.timeout_seconds
        self.logger.info(f"Export timeout after {timeout}s")

        # Check if any devices incomplete
        incomplete = [
            sn for sn in self.discovered_devices if not self._is_device_complete(sn)
        ]

        if incomplete:
            self.logger.warning(f"Partial export: {len(incomplete)} incomplete devices")
            self.export_status = "FAILED_TIMEOUT"

        self._finalize_export()

    def failed(self, message: str) -> None:
        """
        Handle connection failure event.

        Args:
            message: Failure message.
        """
        self.logger.error(f"Connection failed: {message}")
        self.export_status = "FAILED_CONNECTION"
        self.export_result.success = False
        self.export_result.error = message
        self.export_result.export_status = self.export_status
        self.on_finish.emit(self.export_result)

    def set_timeout(self, timeout_seconds: float) -> None:
        """
        Set timeout for export operation.

        Args:
            timeout_seconds: Timeout in seconds.
        """
        self.logger.debug(f"Set timeout: {timeout_seconds}s")
        self.conbus_protocol.timeout_seconds = timeout_seconds

    def set_event_loop(self, event_loop: asyncio.AbstractEventLoop) -> None:
        """
        Set event loop for async operations.

        Args:
            event_loop: Event loop to use.
        """
        self.logger.debug("Set event loop")
        self.conbus_protocol.set_event_loop(event_loop)

    def start_reactor(self) -> None:
        """Start the reactor."""
        self.conbus_protocol.start_reactor()

    def stop_reactor(self) -> None:
        """Stop the reactor."""
        self.conbus_protocol.stop_reactor()

    def __enter__(self) -> "ConbusExportService":
        """
        Enter context manager.

        Returns:
            Self for context manager protocol.
        """
        # Reset state for reuse
        self.discovered_devices = []
        self.device_configs = {}
        self.export_result = ConbusExportResponse(success=False)
        self.export_status = "OK"
        self._finalized = False
        return self

    def __exit__(
        self, _exc_type: Optional[type], _exc_val: Optional[Exception], _exc_tb: Any
    ) -> None:
        """Exit context manager and disconnect signals."""
        self.conbus_protocol.on_connection_made.disconnect(self.connection_made)
        self.conbus_protocol.on_telegram_sent.disconnect(self.telegram_sent)
        self.conbus_protocol.on_telegram_received.disconnect(self.telegram_received)
        self.conbus_protocol.on_timeout.disconnect(self.timeout)
        self.conbus_protocol.on_failed.disconnect(self.failed)
        self.on_progress.disconnect()
        self.on_device_exported.disconnect()
        self.on_finish.disconnect()
        self.stop_reactor()



"""Conbus CLI commands package."""

from xp.cli.commands.conbus.conbus import (
    conbus,
    conbus_actiontable,
    conbus_autoreport,
    conbus_blink,
    conbus_datapoint,
    conbus_export,
    conbus_lightlevel,
    conbus_linknumber,
    conbus_msactiontable,
    conbus_output,
)

__all__ = [
    "conbus",
    "conbus_blink",
    "conbus_output",
    "conbus_datapoint",
    "conbus_linknumber",
    "conbus_autoreport",
    "conbus_lightlevel",
    "conbus_msactiontable",
    "conbus_actiontable",
    "conbus_export",
]



"""Unit tests for HomekitService."""

from unittest.mock import Mock

import pytest

from xp.models.config.conson_module_config import (
    ConsonModuleConfig,
    ConsonModuleListConfig,
)
from xp.models.homekit.homekit_config import (
    BridgeConfig,
    HomekitAccessoryConfig,
    HomekitConfig,
    RoomConfig,
)
from xp.models.protocol.conbus_protocol import TelegramReceivedEvent
from xp.models.telegram.datapoint_type import DataPointType
from xp.models.telegram.event_telegram import EventTelegram
from xp.models.telegram.event_type import EventType
from xp.models.telegram.reply_telegram import ReplyTelegram
from xp.models.telegram.system_function import SystemFunction
from xp.models.term.connection_state import ConnectionState
from xp.services.term.homekit_service import HomekitService


class TestHomekitService:
    """Unit tests for HomekitService."""

    def _make_event(
        self,
        mock_protocol: Mock,
        frame: str,
        telegram_type: str,
        serial_number: str = "",
    ) -> TelegramReceivedEvent:
        """Create TelegramReceivedEvent helper."""
        telegram = frame[1:-1]
        checksum = telegram[-2:]
        payload = telegram[:-2]
        return TelegramReceivedEvent(
            protocol=mock_protocol,
            frame=frame,
            telegram=telegram,
            payload=payload,
            telegram_type=telegram_type,
            serial_number=serial_number,
            checksum=checksum,
            checksum_valid=True,
        )

    @pytest.fixture
    def mock_protocol(self):
        """Create mock ConbusEventProtocol."""
        from xp.services.protocol.conbus_event_protocol import ConbusEventProtocol

        protocol = Mock(spec=ConbusEventProtocol)
        protocol.cli_config = Mock()
        protocol.cli_config.ip = "192.168.1.100"
        protocol.cli_config.port = 10001
        protocol.on_connection_made = Mock()
        protocol.on_connection_failed = Mock()
        protocol.on_telegram_received = Mock()
        protocol.on_timeout = Mock()
        protocol.on_failed = Mock()
        protocol.connect = Mock()
        protocol.disconnect = Mock()
        protocol.send_telegram = Mock()
        protocol.send_raw_telegram = Mock()
        return protocol

    @pytest.fixture
    def homekit_config(self):
        """Create HomekitConfig with test accessories."""
        return HomekitConfig(
            bridge=BridgeConfig(
                name="Test Bridge",
                rooms=[
                    RoomConfig(name="Living Room", accessories=["light1", "dimmer1"]),
                    RoomConfig(name="Bedroom", accessories=["light2"]),
                ],
            ),
            accessories=[
                HomekitAccessoryConfig(
                    name="light1",
                    id="light1",
                    serial_number="1234567890",
                    output_number=0,
                    description="Main Light",
                    service="Lightbulb",
                    on_action="E02L09I80",
                    off_action="E02L09I80",
                    toggle_action="E02L09I00",
                ),
                HomekitAccessoryConfig(
                    name="dimmer1",
                    id="dimmer1",
                    serial_number="2222222222",
                    output_number=0,
                    description="Dimmer Light",
                    service="Lightbulb",
                    on_action="E02L03I80",
                    off_action="E02L03I80",
                    toggle_action="E02L03I00",
                ),
                HomekitAccessoryConfig(
                    name="light2",
                    id="light2",
                    serial_number="1234567890",
                    output_number=1,
                    description="Bedroom Light",
                    service="Lightbulb",
                    on_action="E02L09I81",
                    off_action="E02L09I81",
                    toggle_action="E02L09I01",
                ),
            ],
        )

    @pytest.fixture
    def conson_config(self):
        """Create ConsonModuleListConfig with test modules."""
        return ConsonModuleListConfig(
            root=[
                ConsonModuleConfig(
                    name="A01",
                    serial_number="1234567890",
                    module_type="XP24",
                    module_type_code=7,
                    link_number=9,
                    auto_report_status="PP",
                ),
                ConsonModuleConfig(
                    name="A02",
                    serial_number="2222222222",
                    module_type="XP33LED",
                    module_type_code=35,
                    link_number=3,
                    auto_report_status="PP",
                ),
            ]
        )

    @pytest.fixture
    def mock_telegram_service(self):
        """Create mock TelegramService."""
        service = Mock()
        service.parse_event_telegram = Mock()
        service.parse_reply_telegram = Mock()
        return service

    @pytest.fixture
    def mock_accessory_driver(self):
        """Create mock HomekitAccessoryDriver."""
        from xp.services.term.homekit_accessory_driver import HomekitAccessoryDriver

        driver = Mock(spec=HomekitAccessoryDriver)
        driver.set_callback = Mock()
        driver.start = Mock()
        driver.stop = Mock()
        driver.update_state = Mock()
        return driver

    @pytest.fixture
    def service(
        self,
        mock_protocol,
        homekit_config,
        conson_config,
        mock_telegram_service,
        mock_accessory_driver,
    ):
        """Create service instance."""
        return HomekitService(
            conbus_protocol=mock_protocol,
            homekit_config=homekit_config,
            conson_config=conson_config,
            telegram_service=mock_telegram_service,
            accessory_driver=mock_accessory_driver,
        )

    def test_initialization(self, service):
        """Test service initializes accessory states from config."""
        assert len(service.accessory_states) == 3

        # Check first accessory (XP24)
        light1 = next(
            (a for a in service.accessory_states if a.accessory_name == "Main Light"),
            None,
        )
        assert light1 is not None
        assert light1.room_name == "Living Room"
        assert light1.serial_number == "1234567890"
        assert light1.module_type == "XP24"
        assert light1.module_name == "A01"
        assert light1.output == 1
        assert light1.output_state == "?"
        assert light1.error_status == "OK"
        assert light1.toggle_action == "E02L09I00"

    def test_initialization_assigns_action_keys(self, service):
        """Test service assigns sequential action keys to accessories."""
        states = service.accessory_states
        actions = [s.action for s in states]
        assert "a" in actions
        assert "b" in actions
        assert "c" in actions

    def test_initialization_assigns_sort_order(self, service):
        """Test service assigns sort order matching config order."""
        states = service.accessory_states
        # Should be sorted by sort field
        assert states[0].sort < states[1].sort < states[2].sort

    def test_connection_state_property(self, service):
        """Test connection_state returns current state."""
        assert service.connection_state == ConnectionState.DISCONNECTED

    def test_server_info_property(self, service):
        """Test server_info returns IP:port."""
        assert service.server_info == "192.168.1.100:10001"

    def test_connect(self, service, mock_protocol):
        """Test connect initiates connection."""
        signal_handler = Mock()
        service.on_connection_state_changed.connect(signal_handler)

        service.connect()

        mock_protocol.connect.assert_called_once()
        assert service.connection_state == ConnectionState.CONNECTING
        signal_handler.assert_called_with(ConnectionState.CONNECTING)

    def test_connect_when_already_connecting(self, service, mock_protocol):
        """Test connect does nothing when already connecting."""
        service.connect()  # First connect
        mock_protocol.connect.reset_mock()

        service.connect()  # Second connect should be ignored

        mock_protocol.connect.assert_not_called()

    def test_disconnect(self, service, mock_protocol):
        """Test disconnect terminates connection."""
        # First connect and establish connection
        service.connect()
        service._on_connection_made()

        signal_handler = Mock()
        service.on_connection_state_changed.connect(signal_handler)

        service.disconnect()

        mock_protocol.disconnect.assert_called_once()
        assert service.connection_state == ConnectionState.DISCONNECTED

    def test_disconnect_when_disconnected(self, service, mock_protocol):
        """Test disconnect does nothing when already disconnected."""
        service.disconnect()

        mock_protocol.disconnect.assert_not_called()

    def test_toggle_connection_connects_when_disconnected(self, service, mock_protocol):
        """Test toggle_connection connects when disconnected."""
        service.toggle_connection()

        mock_protocol.connect.assert_called_once()
        assert service.connection_state == ConnectionState.CONNECTING

    def test_toggle_connection_disconnects_when_connected(self, service, mock_protocol):
        """Test toggle_connection disconnects when connected."""
        service.connect()
        service._on_connection_made()

        service.toggle_connection()

        mock_protocol.disconnect.assert_called_once()

    def test_select_accessory_returns_id(self, service):
        """Test select_accessory returns accessory ID for valid key."""
        accessory_id = service.select_accessory("a")

        assert accessory_id is not None
        assert accessory_id == "A01_1"

    def test_select_accessory_invalid_key(self, service):
        """Test select_accessory returns None for invalid key."""
        accessory_id = service.select_accessory("z")  # Not assigned

        assert accessory_id is None

    def test_toggle_selected_sends_telegram(self, service, mock_protocol):
        """Test toggle_selected sends toggle_action telegram."""
        service.connect()
        service._on_connection_made()

        accessory_id = service.select_accessory("a")
        result = service.toggle_selected(accessory_id)

        assert result is True
        mock_protocol.send_raw_telegram.assert_called()

    def test_toggle_selected_invalid_id(self, service, mock_protocol):
        """Test toggle_selected returns False for invalid ID."""
        result = service.toggle_selected("invalid_id")

        assert result is False

    def test_turn_on_selected_sends_telegram(self, service, mock_protocol):
        """Test turn_on_selected sends on_action telegram."""
        accessory_id = service.select_accessory("a")
        result = service.turn_on_selected(accessory_id)

        assert result is True
        mock_protocol.send_raw_telegram.assert_called()

    def test_turn_off_selected_sends_telegram(self, service, mock_protocol):
        """Test turn_off_selected sends off_action telegram."""
        accessory_id = service.select_accessory("a")
        result = service.turn_off_selected(accessory_id)

        assert result is True
        mock_protocol.send_raw_telegram.assert_called()

    def test_toggle_selected_no_toggle_action(
        self, mock_protocol, conson_config, mock_telegram_service, mock_accessory_driver
    ):
        """Test toggle_selected returns False when no toggle_action."""
        # Create config without toggle_action
        homekit_config = HomekitConfig(
            bridge=BridgeConfig(
                name="Test",
                rooms=[RoomConfig(name="Room", accessories=["light_no_toggle"])],
            ),
            accessories=[
                HomekitAccessoryConfig(
                    name="light_no_toggle",
                    id="light_no_toggle",
                    serial_number="1234567890",
                    output_number=0,
                    description="No Toggle Light",
                    service="Lightbulb",
                    on_action="E02L09I80",
                    off_action="E02L09I80",
                    toggle_action=None,
                ),
            ],
        )

        svc = HomekitService(
            conbus_protocol=mock_protocol,
            homekit_config=homekit_config,
            conson_config=conson_config,
            telegram_service=mock_telegram_service,
            accessory_driver=mock_accessory_driver,
        )
        accessory_id = svc.select_accessory("a")
        assert accessory_id is not None
        result = svc.toggle_selected(accessory_id)

        assert result is False

    def test_refresh_all_queries_eligible_modules(self, service, mock_protocol):
        """Test refresh_all queries XP24/XP33LR/XP33LED modules."""
        service.connect()
        service._on_connection_made()

        service.refresh_all()

        # Should query both modules (XP24 and XP33LED)
        assert mock_protocol.send_telegram.call_count == 2

    def test_on_connection_made_handler(self, service):
        """Test _on_connection_made updates state and emits signals."""
        service.connect()

        signal_handler = Mock()
        service.on_connection_state_changed.connect(signal_handler)

        room_list_handler = Mock()
        service.on_room_list_updated.connect(room_list_handler)

        service._on_connection_made()

        assert service.connection_state == ConnectionState.CONNECTED
        signal_handler.assert_called_with(ConnectionState.CONNECTED)
        room_list_handler.assert_called_once()

    def test_on_connection_failed_handler(self, service):
        """Test _on_connection_failed updates state."""
        service.connect()

        signal_handler = Mock()
        service.on_connection_state_changed.connect(signal_handler)

        service._on_connection_failed(Exception("Connection error"))

        assert service.connection_state == ConnectionState.FAILED
        signal_handler.assert_called_with(ConnectionState.FAILED)

    def test_on_timeout_handler(self, service):
        """Test _on_timeout emits status message."""
        status_handler = Mock()
        service.on_status_message.connect(status_handler)

        service._on_timeout()

        status_handler.assert_called_once()

    def test_on_failed_handler(self, service):
        """Test _on_failed updates state."""
        service.connect()

        signal_handler = Mock()
        service.on_connection_state_changed.connect(signal_handler)

        service._on_failed(Exception("Protocol error"))

        assert service.connection_state == ConnectionState.FAILED

    def test_handle_event_telegram_xp24_output_on(
        self, service, mock_telegram_service, mock_protocol
    ):
        """Test _handle_event_telegram processes XP24 output ON event."""
        event_telegram = EventTelegram(
            raw_telegram="<E07L09I80MAE>",
            checksum="AE",
            checksum_validated=True,
            event_telegram_type="E",
            module_type=7,  # XP24
            link_number=9,
            input_number=80,  # Output 0
            event_type=EventType.BUTTON_PRESS,
        )
        mock_telegram_service.parse_event_telegram.return_value = event_telegram

        signal_handler = Mock()
        service.on_module_state_changed.connect(signal_handler)

        event = self._make_event(mock_protocol, "<E07L09I80MAE>", "E")
        service._on_telegram_received(event)

        # Find the accessory with output=1 for module A01
        light1 = next(
            (a for a in service.accessory_states if a.accessory_name == "Main Light"),
            None,
        )
        assert light1 is not None
        assert light1.output_state == "ON"
        assert light1.last_update is not None
        signal_handler.assert_called()

    def test_handle_event_telegram_xp24_output_off(
        self, service, mock_telegram_service, mock_protocol
    ):
        """Test _handle_event_telegram processes XP24 output OFF event."""
        event_telegram = EventTelegram(
            raw_telegram="<E07L09I80BAE>",
            checksum="AE",
            checksum_validated=True,
            event_telegram_type="E",
            module_type=7,  # XP24
            link_number=9,
            input_number=80,  # Output 0
            event_type=EventType.BUTTON_RELEASE,
        )
        mock_telegram_service.parse_event_telegram.return_value = event_telegram

        event = self._make_event(mock_protocol, "<E07L09I80BAE>", "E")
        service._on_telegram_received(event)

        light1 = next(
            (a for a in service.accessory_states if a.accessory_name == "Main Light"),
            None,
        )
        assert light1 is not None
        assert light1.output_state == "OFF"

    def test_handle_event_telegram_xp33led_channel_on(
        self, service, mock_telegram_service, mock_protocol
    ):
        """Test _handle_event_telegram processes XP33LED channel ON event."""
        event_telegram = EventTelegram(
            raw_telegram="<E35L03I80MAE>",
            checksum="AE",
            checksum_validated=True,
            event_telegram_type="E",
            module_type=35,  # XP33LED
            link_number=3,
            input_number=80,  # Channel 0
            event_type=EventType.BUTTON_PRESS,
        )
        mock_telegram_service.parse_event_telegram.return_value = event_telegram

        event = self._make_event(mock_protocol, "<E35L03I80MAE>", "E")
        service._on_telegram_received(event)

        dimmer = next(
            (a for a in service.accessory_states if a.accessory_name == "Dimmer Light"),
            None,
        )
        assert dimmer is not None
        assert dimmer.output_state == "ON"
        assert dimmer.dimming_state == ""  # Empty when ON

    def test_handle_event_telegram_xp33led_channel_off(
        self, service, mock_telegram_service, mock_protocol
    ):
        """Test _handle_event_telegram processes XP33LED channel OFF event."""
        event_telegram = EventTelegram(
            raw_telegram="<E35L03I80BAE>",
            checksum="AE",
            checksum_validated=True,
            event_telegram_type="E",
            module_type=35,  # XP33LED
            link_number=3,
            input_number=80,  # Channel 0
            event_type=EventType.BUTTON_RELEASE,
        )
        mock_telegram_service.parse_event_telegram.return_value = event_telegram

        event = self._make_event(mock_protocol, "<E35L03I80BAE>", "E")
        service._on_telegram_received(event)

        dimmer = next(
            (a for a in service.accessory_states if a.accessory_name == "Dimmer Light"),
            None,
        )
        assert dimmer is not None
        assert dimmer.output_state == "OFF"
        assert dimmer.dimming_state == "-"  # Dash when OFF and dimmable

    def test_handle_event_telegram_ignores_input_events(
        self, service, mock_telegram_service, mock_protocol
    ):
        """Test _handle_event_telegram ignores input events (I00-I09)."""
        event_telegram = EventTelegram(
            raw_telegram="<E07L09I02MAE>",
            checksum="AE",
            checksum_validated=True,
            module_type=7,
            link_number=9,
            input_number=2,  # Input event, not output
            event_type=EventType.BUTTON_PRESS,
        )
        mock_telegram_service.parse_event_telegram.return_value = event_telegram

        signal_handler = Mock()
        service.on_module_state_changed.connect(signal_handler)

        event = self._make_event(mock_protocol, "<E07L09I02MAE>", "E")
        service._on_telegram_received(event)

        # Signal should not be emitted for input events
        signal_handler.assert_not_called()

    def test_handle_event_telegram_malformed(
        self, service, mock_telegram_service, mock_protocol
    ):
        """Test _handle_event_telegram handles malformed telegram."""
        mock_telegram_service.parse_event_telegram.return_value = None

        event = self._make_event(mock_protocol, "<INVALID>", "E")

        # Should not raise exception
        service._on_telegram_received(event)

    def test_handle_reply_telegram_updates_outputs(
        self, service, mock_telegram_service, mock_protocol
    ):
        """Test _handle_reply_telegram updates accessory outputs."""
        reply_telegram = ReplyTelegram(
            raw_telegram="<R1234567890F42D0009AB>",
            checksum="AB",
            checksum_validated=True,
            serial_number="1234567890",
            system_function=SystemFunction.READ_DATAPOINT,
            datapoint_type=DataPointType.MODULE_OUTPUT_STATE,
            data_value="03",  # Binary 0011 = outputs 0,1 ON
        )
        mock_telegram_service.parse_reply_telegram.return_value = reply_telegram

        signal_handler = Mock()
        service.on_module_state_changed.connect(signal_handler)

        event = TelegramReceivedEvent(
            protocol=mock_protocol,
            frame="<R1234567890F42D0009AB>",
            telegram="R1234567890F42D0009AB",
            payload="R1234567890F42D0009",
            telegram_type="R",
            serial_number="1234567890",
            checksum="AB",
            checksum_valid=True,
        )

        service._on_telegram_received(event)

        # Both accessories with serial 1234567890 should be updated
        light1 = next(
            (a for a in service.accessory_states if a.accessory_name == "Main Light"),
            None,
        )
        light2 = next(
            (
                a
                for a in service.accessory_states
                if a.accessory_name == "Bedroom Light"
            ),
            None,
        )

        assert light1 is not None
        assert light2 is not None
        assert signal_handler.call_count == 2  # Called for each accessory

    def test_handle_reply_telegram_no_serial(
        self, service, mock_telegram_service, mock_protocol
    ):
        """Test _handle_reply_telegram ignores events without serial number."""
        event = TelegramReceivedEvent(
            protocol=mock_protocol,
            frame="<RAB>",
            telegram="RAB",
            payload="R",
            telegram_type="R",
            serial_number="",  # No serial
            checksum="AB",
            checksum_valid=True,
        )

        # Should not raise exception
        service._on_telegram_received(event)

    def test_handle_reply_telegram_not_output_state(
        self, service, mock_telegram_service, mock_protocol
    ):
        """Test _handle_reply_telegram ignores non-output-state replies."""
        reply_telegram = ReplyTelegram(
            raw_telegram="<R1234567890F01D00AB>",
            checksum="AB",
            checksum_validated=True,
            serial_number="1234567890",
            system_function=SystemFunction.DISCOVERY,  # Not READ_DATAPOINT
            datapoint_type=None,
            data_value="",
        )
        mock_telegram_service.parse_reply_telegram.return_value = reply_telegram

        signal_handler = Mock()
        service.on_module_state_changed.connect(signal_handler)

        event = TelegramReceivedEvent(
            protocol=mock_protocol,
            frame="<R1234567890F01D00AB>",
            telegram="R1234567890F01D00AB",
            payload="R1234567890F01D00",
            telegram_type="R",
            serial_number="1234567890",
            checksum="AB",
            checksum_valid=True,
        )

        service._on_telegram_received(event)

        signal_handler.assert_not_called()

    def test_cleanup(self, service, mock_protocol):
        """Test cleanup disconnects signals."""
        service.cleanup()

        mock_protocol.on_connection_made.disconnect.assert_called_once()
        mock_protocol.on_connection_failed.disconnect.assert_called_once()
        mock_protocol.on_telegram_received.disconnect.assert_called_once()
        mock_protocol.on_timeout.disconnect.assert_called_once()
        mock_protocol.on_failed.disconnect.assert_called_once()

    def test_context_manager(self, service, mock_protocol):
        """Test context manager entry and exit."""
        with service as svc:
            assert svc is service

        # Verify cleanup was called
        mock_protocol.on_connection_made.disconnect.assert_called_once()

    def test_accessory_not_in_config(
        self, mock_protocol, conson_config, mock_telegram_service, mock_accessory_driver
    ):
        """Test service handles missing accessory config gracefully."""
        homekit_config = HomekitConfig(
            bridge=BridgeConfig(
                name="Test",
                rooms=[RoomConfig(name="Room", accessories=["nonexistent"])],
            ),
            accessories=[],  # No accessories defined
        )

        service = HomekitService(
            conbus_protocol=mock_protocol,
            homekit_config=homekit_config,
            conson_config=conson_config,
            telegram_service=mock_telegram_service,
            accessory_driver=mock_accessory_driver,
        )

        # Should handle gracefully with empty states
        assert len(service.accessory_states) == 0

    def test_module_not_in_config(
        self, mock_protocol, mock_telegram_service, mock_accessory_driver
    ):
        """Test service handles missing module config gracefully."""
        homekit_config = HomekitConfig(
            bridge=BridgeConfig(
                name="Test",
                rooms=[RoomConfig(name="Room", accessories=["light"])],
            ),
            accessories=[
                HomekitAccessoryConfig(
                    name="light",
                    id="light",
                    serial_number="9999999999",  # Not in conson_config
                    output_number=0,
                    description="Light",
                    service="Lightbulb",
                    on_action="E02L09I80",
                    off_action="E02L09I80",
                ),
            ],
        )

        conson_config = ConsonModuleListConfig(root=[])  # Empty config

        service = HomekitService(
            conbus_protocol=mock_protocol,
            homekit_config=homekit_config,
            conson_config=conson_config,
            telegram_service=mock_telegram_service,
            accessory_driver=mock_accessory_driver,
        )

        # Should handle gracefully with empty states
        assert len(service.accessory_states) == 0



"""Integration tests for module command functionality."""

import json

from click.testing import CliRunner

from xp.cli.main import cli


class TestModuleIntegration:
    """Integration tests for module command functionality."""

    def setup_method(self):
        """Set up test fixtures."""
        self.runner = CliRunner()

    def test_module_info_command_by_code(self):
        """Test that module info command with code works."""
        result = self.runner.invoke(cli, ["module", "info", "14"])

        assert result.exit_code == 0

        # Parse JSON output
        output = json.loads(result.output)
        assert output["code"] == 14
        assert output["name"] == "XP2606"
        assert "5 way push button panel" in output["description"]
        assert output["category"] == "Interface Panels"

    def test_module_info_command_by_name(self):
        """Test module info command with name."""
        result = self.runner.invoke(cli, ["module", "info", "XP2606"])

        assert result.exit_code == 0

        # Parse JSON output
        output = json.loads(result.output)
        assert output["code"] == 14
        assert output["name"] == "XP2606"
        assert output["is_push_button_panel"] is True

    def test_module_info_command_json_output(self):
        """Test module info command with JSON output."""
        result = self.runner.invoke(cli, ["module", "info", "14"])

        assert result.exit_code == 0

        output = json.loads(result.output)
        assert output["code"] == 14
        assert output["name"] == "XP2606"
        assert output["category"] == "Interface Panels"
        assert output["is_push_button_panel"] is True

    def test_module_info_command_invalid_code(self):
        """Test module info command with invalid code."""
        result = self.runner.invoke(cli, ["module", "info", "999"])

        assert result.exit_code == 1

        # Parse JSON error response
        output = json.loads(result.output)
        assert output["success"] is False
        assert "error" in output
        assert "Module type with code 999 not found" in output["error"]

    def test_module_info_command_invalid_code_json(self):
        """Test module info command with invalid code and JSON output."""
        result = self.runner.invoke(cli, ["module", "info", "999"])

        assert result.exit_code == 1

        output = json.loads(result.output)
        assert output["success"] is False
        assert "error" in output
        assert "999" in output["error"]

    def test_module_list_command(self):
        """Test module list command."""
        result = self.runner.invoke(cli, ["module", "list"])

        assert result.exit_code == 0

        # Parse JSON output
        output = json.loads(result.output)
        assert "modules" in output
        assert "count" in output
        # Check for specific modules
        module_names = [m["name"] for m in output["modules"]]
        assert "NOMOD" in module_names
        assert "XP2606" in module_names

    def test_module_list_command_json_output(self):
        """Test module list command with JSON output."""
        result = self.runner.invoke(cli, ["module", "list"])

        assert result.exit_code == 0

        output = json.loads(result.output)
        assert "modules" in output
        assert "count" in output
        assert output["count"] == 37
        assert len(output["modules"]) == 37

    def test_module_list_command_by_category(self):
        """Test module list command filtered by category."""
        result = self.runner.invoke(cli, ["module", "list", "--category", "System"])

        assert result.exit_code == 0

        # Parse JSON output
        output = json.loads(result.output)
        assert "modules" in output
        # Check for specific modules in System category
        module_names = [m["name"] for m in output["modules"]]
        assert "NOMOD" in module_names
        assert "ALLMOD" in module_names

    def test_module_list_command_group_by_category(self):
        """Test module list command grouped by category."""
        result = self.runner.invoke(cli, ["module", "list", "--group-by-category"])

        assert result.exit_code == 0

        # Parse JSON output
        output = json.loads(result.output)
        assert "modules_by_category" in output
        assert "System" in output["modules_by_category"]
        assert "CP Link Modules" in output["modules_by_category"]
        assert "XP Control Modules" in output["modules_by_category"]
        assert "Interface Panels" in output["modules_by_category"]

    def test_module_list_command_group_by_category_json(self):
        """Test module list command grouped by category with JSON output."""
        result = self.runner.invoke(cli, ["module", "list", "--group-by-category"])

        assert result.exit_code == 0

        output = json.loads(result.output)
        assert "modules_by_category" in output
        assert "System" in output["modules_by_category"]
        assert "Interface Panels" in output["modules_by_category"]

    def test_module_list_command_invalid_category(self):
        """Test module list command with invalid category."""
        result = self.runner.invoke(cli, ["module", "list", "--category", "Invalid"])

        assert result.exit_code == 0

        # Parse JSON output
        output = json.loads(result.output)
        assert "modules" in output
        assert len(output["modules"]) == 0
        assert output["category"] == "Invalid"

    def test_module_search_command(self):
        """Test module search command."""
        result = self.runner.invoke(cli, ["module", "search", "push button"])

        assert result.exit_code == 0

        # Parse JSON output
        output = json.loads(result.output)
        assert output["query"] == "push button"
        assert output["count"] >= 1
        assert "matches" in output
        # Check that XP2606 is in the search results
        match_names = [m["name"] for m in output["matches"]]
        assert "XP2606" in match_names

    def test_module_search_command_json_output(self):
        """Test module search command with JSON output."""
        result = self.runner.invoke(cli, ["module", "search", "XP2606"])

        assert result.exit_code == 0

        output = json.loads(result.output)
        assert output["query"] == "XP2606"
        assert output["count"] >= 1
        assert "matches" in output
        assert any(match["name"] == "XP2606" for match in output["matches"])

    def test_module_search_command_by_name_field(self):
        """Test module search command searching only name field."""
        result = self.runner.invoke(cli, ["module", "search", "XP", "--field", "name"])

        assert result.exit_code == 0
        assert "XP2606" in result.output
        assert "XP24" in result.output

    def test_module_search_command_no_matches(self):
        """Test module search command with no matches."""
        result = self.runner.invoke(cli, ["module", "search", "NONEXISTENT"])

        assert result.exit_code == 0

        # Parse JSON output
        output = json.loads(result.output)
        assert output["query"] == "NONEXISTENT"
        assert output["count"] == 0
        assert len(output["matches"]) == 0

    def test_module_categories_command(self):
        """Test module categories command."""
        result = self.runner.invoke(cli, ["module", "categories"])

        assert result.exit_code == 0

        # Parse JSON output
        output = json.loads(result.output)
        assert "categories" in output
        assert "System" in output["categories"]
        assert "Interface Panels" in output["categories"]

    def test_module_categories_command_json_output(self):
        """Test module categories command with JSON output."""
        result = self.runner.invoke(cli, ["module", "categories"])

        assert result.exit_code == 0

        output = json.loads(result.output)
        assert "categories" in output
        assert "System" in output["categories"]
        assert "Interface Panels" in output["categories"]
        assert output["categories"]["System"] == 2  # NOMOD and ALLMOD

    def test_module_help_command(self):
        """Test module help command."""
        result = self.runner.invoke(cli, ["module", "--help"])

        assert result.exit_code == 0
        assert "module type operations" in result.output
        assert "info" in result.output
        assert "list" in result.output
        assert "search" in result.output
        assert "categories" in result.output

    def test_module_subcommand_help(self):
        """Test module subcommand help."""
        result = self.runner.invoke(cli, ["module", "info", "--help"])

        assert result.exit_code == 0
        assert "Get information about a module type" in result.output
        assert "Examples:" in result.output
        assert "xp module info 14" in result.output

    def test_enhanced_telegram_parsing_with_module_info(self):
        """Test that telegram parsing now includes module information."""
        result = self.runner.invoke(cli, ["telegram", "parse", "<E14L00I02MAK>"])

        assert result.exit_code == 0

        output = json.loads(result.output)
        assert output["module_type"] == 14
        assert "module_info" in output
        assert output["module_info"]["name"] == "XP2606"
        assert (
            output["module_info"]["description"]
            == "5 way push button panel with sesam, L-Team design"
        )
        assert output["module_info"]["category"] == "Interface Panels"

    def test_enhanced_telegram_parsing_human_readable(self):
        """Test that telegram parsing includes module names in JSON."""
        result = self.runner.invoke(cli, ["telegram", "parse", "<E14L00I02MAK>"])

        assert result.exit_code == 0

        # Parse JSON output
        output = json.loads(result.output)
        assert output["module_type"] == 14
        assert "module_info" in output
        assert output["module_info"]["name"] == "XP2606"
        assert output["event_type_name"] == "button_press"

    def test_enhanced_telegram_parsing_unknown_module(self):
        """Test telegram parsing with unknown module type."""
        # Test with a module type that doesn't exist (using high number)
        # This tests the graceful handling when module_info is None
        result = self.runner.invoke(cli, ["telegram", "parse", "<E99L00I02MAK>"])

        assert result.exit_code == 0

        output = json.loads(result.output)
        assert output["module_type"] == 99
        assert output["module_info"] is None



"""Conbus client operations CLI commands."""

import json

import click
from click import Context

from xp.cli.commands.conbus.conbus import conbus_blink
from xp.cli.utils.decorators import (
    connection_command,
    handle_service_errors,
)
from xp.cli.utils.serial_number_type import SERIAL
from xp.models.conbus.conbus_blink import ConbusBlinkResponse
from xp.services.conbus.conbus_blink_all_service import ConbusBlinkAllService
from xp.services.conbus.conbus_blink_service import ConbusBlinkService
from xp.services.telegram.telegram_blink_service import BlinkError


@conbus_blink.command("on", short_help="Blink on remote service")
@click.argument("serial_number", type=SERIAL)
@click.pass_context
@connection_command()
@handle_service_errors(BlinkError)
def send_blink_on_telegram(ctx: Context, serial_number: str) -> None:
    r"""
    Send blink command to start blinking module LED.

    Args:
        ctx: Click context object.
        serial_number: 10-digit module serial number.

    Examples:
        \b
        xp conbus blink on 0012345008
    """

    def on_finish(service_response: ConbusBlinkResponse) -> None:
        """
        Handle successful completion of blink on command.

        Args:
            service_response: Blink response object.
        """
        click.echo(json.dumps(service_response.to_dict(), indent=2))
        service.stop_reactor()

    service: ConbusBlinkService = (
        ctx.obj.get("container").get_container().resolve(ConbusBlinkService)
    )
    with service:
        service.on_finish.connect(on_finish)
        service.send_blink_telegram(serial_number, "on", 0.5)
        service.start_reactor()


@conbus_blink.command("off")
@click.argument("serial_number", type=SERIAL)
@click.pass_context
@connection_command()
@handle_service_errors(BlinkError)
def send_blink_off_telegram(ctx: Context, serial_number: str) -> None:
    r"""
    Send blink command to stop blinking module LED.

    Args:
        ctx: Click context object.
        serial_number: 10-digit module serial number.

    Examples:
        \b
        xp conbus blink off 0012345008
    """

    def on_finish(service_response: ConbusBlinkResponse) -> None:
        """
        Handle successful completion of blink off command.

        Args:
            service_response: Blink response object.
        """
        click.echo(json.dumps(service_response.to_dict(), indent=2))
        service.stop_reactor()

    service: ConbusBlinkService = (
        ctx.obj.get("container").get_container().resolve(ConbusBlinkService)
    )
    with service:
        service.on_finish.connect(on_finish)
        service.send_blink_telegram(serial_number, "off", 0.5)
        service.start_reactor()


@conbus_blink.group("all", short_help="Control blink state for all devices")
def conbus_blink_all() -> None:
    """Control blink state for all discovered devices."""
    pass


@conbus_blink_all.command("off", short_help="Turn off blinking for all devices")
@click.pass_context
@connection_command()
@handle_service_errors(BlinkError)
def blink_all_off(ctx: Context) -> None:
    r"""
    Turn off blinking for all discovered devices.

    Args:
        ctx: Click context object.

    Examples:
        \b
        xp conbus blink all off
    """

    def on_finish(discovered_devices: ConbusBlinkResponse) -> None:
        """
        Handle successful completion of blink all off command.

        Args:
            discovered_devices: Blink response with all devices.
        """
        click.echo(json.dumps(discovered_devices.to_dict(), indent=2))
        service.stop_reactor()

    def progress(message: str) -> None:
        """
        Handle progress updates during blink all off operation.

        Args:
            message: Progress message string.
        """
        click.echo(message, nl=False)

    service: ConbusBlinkAllService = (
        ctx.obj.get("container").get_container().resolve(ConbusBlinkAllService)
    )
    with service:
        service.on_progress.connect(progress)
        service.on_finish.connect(on_finish)
        service.send_blink_all_telegram("off", 5)
        service.start_reactor()


@conbus_blink_all.command("on", short_help="Turn on blinking for all devices")
@click.pass_context
@connection_command()
@handle_service_errors(BlinkError)
def blink_all_on(ctx: Context) -> None:
    r"""
    Turn on blinking for all discovered devices.

    Args:
        ctx: Click context object.

    Examples:
        \b
        xp conbus blink all on
    """

    def on_finish(discovered_devices: ConbusBlinkResponse) -> None:
        """
        Handle successful completion of blink all on command.

        Args:
            discovered_devices: Blink response with all devices.
        """
        click.echo(json.dumps(discovered_devices.to_dict(), indent=2))
        service.stop_reactor()

    def progress(message: str) -> None:
        """
        Handle progress updates during blink all on operation.

        Args:
            message: Progress message string.
        """
        click.echo(message, nl=False)

    service: ConbusBlinkAllService = (
        ctx.obj.get("container").get_container().resolve(ConbusBlinkAllService)
    )
    with service:
        service.on_progress.connect(progress)
        service.on_finish.connect(on_finish)
        service.send_blink_all_telegram("on", 5)
        service.start_reactor()



"""Tests for conbus models."""

from datetime import datetime

from xp.models.conbus.conbus import ConbusRequest, ConbusResponse


class TestConbusRequest:
    """Test ConbusRequest model."""

    def test_to_dict(self):
        """Test to_dict method."""
        timestamp = datetime(2025, 1, 1, 12, 0, 0)
        result = ConbusRequest(
            serial_number="12345",
            function_code="F",
            data="test_data",
            timestamp=timestamp,
        ).to_dict()
        assert result["serial_number"] == "12345"
        assert result["function_code"] == "F"
        assert result["data"] == "test_data"
        assert "2025-01-01T12:00:00" in result["timestamp"]

    def test_post_init_sets_timestamp(self):
        """Test __post_init__ sets timestamp if None."""
        request = ConbusRequest(
            serial_number="12345", function_code="F", data="test", timestamp=None
        )
        assert request.timestamp is not None
        assert isinstance(request.timestamp, datetime)


class TestConbusResponse:
    """Test ConbusResponse model."""

    def test_post_init_sets_defaults(self):
        """Test __post_init__ sets default values."""
        timestamp = datetime(2025, 1, 1, 12, 0, 0)
        response = ConbusResponse(
            success=True,
            sent_telegrams=["<TEST>"],
            received_telegrams=["<REPLY>"],
            timestamp=timestamp,
        )
        assert response.sent_telegrams == []
        assert response.received_telegrams == []
        assert response.error == ""

    def test_to_dict(self):
        """Test to_dict method."""
        timestamp = datetime(2025, 1, 1, 12, 0, 0)
        result = ConbusResponse(
            success=True,
            sent_telegrams=["<TEST>"],
            received_telegrams=["<REPLY>"],
            error="test error",
            timestamp=timestamp,
        ).to_dict()

        assert result["success"] is True
        assert result["sent_telegrams"] == []  # post_init resets these
        assert result["received_telegrams"] == []  # post_init resets these
        assert result["error"] == "test error"
        assert "T" in result["timestamp"]  # ISO format check



"""Integration tests for term command file logging."""

# NOTE: These integration tests have been removed as setup_file_logging
# has been refactored into the LoggerService class.
# Tests need to be rewritten to use the new LoggerService API.


def test_placeholder() -> None:
    """Placeholder test to prevent empty test file errors."""
    pass



"""Integration tests for Conbus datapoint functionality."""

from unittest.mock import Mock

from click.testing import CliRunner

from xp.cli.main import cli
from xp.models.conbus.conbus_datapoint import ConbusDatapointResponse
from xp.models.telegram.system_function import SystemFunction
from xp.services.conbus.conbus_datapoint_queryall_service import (
    ConbusDatapointQueryAllService,
)
from xp.services.conbus.conbus_datapoint_service import (
    ConbusDatapointService,
)


class TestConbusDatapointIntegration:
    """Integration tests for conbus datapoint CLI operations."""

    def setup_method(self):
        """Set up test fixtures."""
        self.runner = CliRunner()
        self.valid_serial = "0123450001"
        self.invalid_serial = "invalid"

    def test_conbus_datapoint_all_valid_serial(self):
        """Test querying all datapoints with valid serial number."""
        # Mock successful response
        mock_service = Mock()
        mock_service.__enter__ = Mock(return_value=mock_service)
        mock_service.__exit__ = Mock(return_value=None)

        mock_response = ConbusDatapointResponse(
            success=True,
            serial_number=self.valid_serial,
            system_function=SystemFunction.READ_DATAPOINT,
            datapoints=[
                {"MODULE_TYPE": "XP33LED"},
                {"HW_VERSION": "XP33LED_HW_VER1"},
                {"SW_VERSION": "XP33LED_V0.04.01"},
                {"AUTO_REPORT_STATUS": "AA"},
                {"MODULE_STATE": "OFF"},
                {"MODULE_OUTPUT_STATE": "xxxxx000"},
            ],
        )

        # Store the callbacks that are connected
        callbacks = {"on_finish": None, "on_progress": None}

        def mock_on_finish_connect(callback):
            """
            Mock on_finish event connection.

            Args:
                callback: Callback function to store.
            """
            callbacks["on_finish"] = callback

        def mock_on_progress_connect(callback):
            """
            Mock on_progress event connection.

            Args:
                callback: Callback function to store.
            """
            callbacks["on_progress"] = callback

        mock_service.on_finish.connect.side_effect = mock_on_finish_connect
        mock_service.on_progress.connect.side_effect = mock_on_progress_connect

        # Make the mock service call the callback immediately
        def mock_query_all_datapoints(serial_number):
            """
            Test helper function.

            Args:
                serial_number: Serial number of the module.
            """
            # Call the on_finish callback that was connected
            if callbacks["on_finish"]:
                callbacks["on_finish"](mock_response)

        def mock_start_reactor() -> None:
            """Mock reactor start method."""
            # Do nothing in test
            pass

        mock_service.query_all_datapoints.side_effect = mock_query_all_datapoints
        mock_service.start_reactor.side_effect = mock_start_reactor

        # Setup mock container to resolve ConbusDatapointQueryAllService
        mock_container = Mock()
        mock_container.resolve.return_value = mock_service
        mock_service_container = Mock()
        mock_service_container.get_container.return_value = mock_container

        # Run CLI command
        result = self.runner.invoke(
            cli,
            ["conbus", "datapoint", "all", self.valid_serial],
            obj={"container": mock_service_container},
        )

        # Debug output
        print(f"Exit code: {result.exit_code}")
        print(f"Output: {result.output}")
        print(f"Exception: {result.exception}")
        print(f"Mock service calls: {mock_service.method_calls}")

        # Assertions
        assert '"success": true' in result.output
        assert result.exit_code == 0
        assert mock_service.query_all_datapoints.called

        # Check the response content
        assert f'"serial_number": "{self.valid_serial}"' in result.output
        assert '"datapoints"' in result.output
        assert '"MODULE_TYPE": "XP33LED"' in result.output

    def test_conbus_datapoint_all_invalid_serial(self):
        """Test querying all datapoints with invalid serial number."""
        # Mock service that raises error
        mock_service = Mock()
        mock_service.__enter__ = Mock(return_value=mock_service)
        mock_service.__exit__ = Mock(return_value=None)

        # Setup mock container to resolve ConbusDatapointService
        mock_container = Mock()
        mock_container.resolve.return_value = mock_service
        mock_service_container = Mock()
        mock_service_container.get_container.return_value = mock_container

        # Run CLI command
        result = self.runner.invoke(
            cli,
            ["conbus", "datapoint", "all", self.invalid_serial],
            obj={"container": mock_service_container},
        )

        # Should handle the error gracefully
        assert result.exit_code != 0
        assert "Invalid serial number" in result.output or "Error" in result.output

    def test_conbus_datapoint_invalid_response(self):
        """Test handling invalid responses from the server."""
        # Mock service with failed response
        mock_service = Mock()
        mock_service.__enter__ = Mock(return_value=mock_service)
        mock_service.__exit__ = Mock(return_value=None)

        mock_response = ConbusDatapointResponse(
            success=False,
            serial_number=self.valid_serial,
            error="Invalid response from server",
            datapoints=[],
        )

        # Store the callbacks that are connected
        callbacks = {"on_finish": None, "on_progress": None}

        def mock_on_finish_connect(callback):
            """
            Mock on_finish event connection.

            Args:
                callback: Callback function to store.
            """
            callbacks["on_finish"] = callback

        def mock_on_progress_connect(callback):
            """
            Mock on_progress event connection.

            Args:
                callback: Callback function to store.
            """
            callbacks["on_progress"] = callback

        mock_service.on_finish.connect.side_effect = mock_on_finish_connect
        mock_service.on_progress.connect.side_effect = mock_on_progress_connect

        # Make the mock service call the callback immediately
        def mock_query_all_datapoints(serial_number):
            """
            Test helper function.

            Args:
                serial_number: Serial number of the module.
            """
            # Call the on_finish callback that was connected
            if callbacks["on_finish"]:
                callbacks["on_finish"](mock_response)

        def mock_start_reactor() -> None:
            """Mock reactor start method."""
            # Do nothing in test
            pass

        mock_service.query_all_datapoints.side_effect = mock_query_all_datapoints
        mock_service.start_reactor.side_effect = mock_start_reactor

        # Setup mock container to resolve ConbusDatapointQueryAllService
        mock_container = Mock()
        mock_container.resolve.return_value = mock_service
        mock_service_container = Mock()
        mock_service_container.get_container.return_value = mock_container

        # Run CLI command
        result = self.runner.invoke(
            cli,
            ["conbus", "datapoint", "all", self.valid_serial],
            obj={"container": mock_service_container},
        )

        # Should return the failed response
        assert '"success": false' in result.output
        assert result.exit_code == 0  # CLI succeeds but response indicates failure
        assert "Invalid response from server" in result.output

    def test_conbus_datapoint_empty_datapoints(self):
        """Test handling when no datapoints are returned."""
        # Mock service with successful but empty response
        mock_service = Mock()
        mock_service.__enter__ = Mock(return_value=mock_service)
        mock_service.__exit__ = Mock(return_value=None)

        mock_response = ConbusDatapointResponse(
            success=True,
            serial_number=self.valid_serial,
            system_function=SystemFunction.READ_DATAPOINT,
            datapoints=[],
        )

        # Store the callbacks that are connected
        callbacks = {"on_finish": None, "on_progress": None}

        def mock_on_finish_connect(callback):
            """
            Mock on_finish event connection.

            Args:
                callback: Callback function to store.
            """
            callbacks["on_finish"] = callback

        def mock_on_progress_connect(callback):
            """
            Mock on_progress event connection.

            Args:
                callback: Callback function to store.
            """
            callbacks["on_progress"] = callback

        mock_service.on_finish.connect.side_effect = mock_on_finish_connect
        mock_service.on_progress.connect.side_effect = mock_on_progress_connect

        # Make the mock service call the callback immediately
        def mock_query_all_datapoints(serial_number):
            """
            Test helper function.

            Args:
                serial_number: Serial number of the module.
            """
            # Call the on_finish callback that was connected
            if callbacks["on_finish"]:
                callbacks["on_finish"](mock_response)

        def mock_start_reactor() -> None:
            """Mock reactor start method."""
            # Do nothing in test
            pass

        mock_service.query_all_datapoints.side_effect = mock_query_all_datapoints
        mock_service.start_reactor.side_effect = mock_start_reactor

        # Setup mock container to resolve ConbusDatapointQueryAllService
        mock_container = Mock()
        mock_container.resolve.return_value = mock_service
        mock_service_container = Mock()
        mock_service_container.get_container.return_value = mock_container

        # Run CLI command
        result = self.runner.invoke(
            cli,
            ["conbus", "datapoint", "all", self.valid_serial],
            obj={"container": mock_service_container},
        )

        # Should succeed with empty datapoints
        assert '"success": true' in result.output
        assert result.exit_code == 0
        assert f'"serial_number": "{self.valid_serial}"' in result.output
        # datapoints field should not be included when empty
        assert '"datapoints"' not in result.output


class TestConbusDatapointService:
    """Unit tests for ConbusDatapointService functionality."""

    def setup_method(self):
        """Set up test fixtures."""
        self.valid_serial = "0123450001"
        self.mock_cli_config = Mock()
        self.mock_reactor = Mock()
        self.mock_telegram_service = Mock()

    def test_service_initialization(self):
        """Test service can be initialized with required dependencies."""
        mock_protocol = Mock()
        service = ConbusDatapointService(
            conbus_protocol=mock_protocol,
            telegram_service=self.mock_telegram_service,
        )

        assert service.telegram_service == self.mock_telegram_service
        assert service.serial_number == ""
        assert service.datapoint_type is None
        assert hasattr(service, "on_finish")
        assert service.service_response.success is False

    def test_service_context_manager(self):
        """Test service can be used as context manager."""
        mock_protocol = Mock()
        service = ConbusDatapointService(
            conbus_protocol=mock_protocol,
            telegram_service=self.mock_telegram_service,
        )

        with service as s:
            assert s is service


class TestConbusDatapointQueryAllService:
    """Unit tests for ConbusDatapointQueryAllService functionality."""

    def setup_method(self):
        """Set up test fixtures."""
        self.valid_serial = "0123450001"
        self.mock_conbus_protocol = Mock()
        self.mock_telegram_service = Mock()

    def test_service_initialization(self):
        """Test service can be initialized with required dependencies."""
        service = ConbusDatapointQueryAllService(
            conbus_protocol=self.mock_conbus_protocol,
            telegram_service=self.mock_telegram_service,
        )

        assert service.telegram_service == self.mock_telegram_service
        assert service.conbus_protocol == self.mock_conbus_protocol

    def test_service_context_manager(self):
        """Test service can be used as context manager."""
        service = ConbusDatapointQueryAllService(
            conbus_protocol=self.mock_conbus_protocol,
            telegram_service=self.mock_telegram_service,
        )

        with service as s:
            assert s is service



"""Event type enumeration for telegram events."""

from enum import Enum


class EventType(Enum):
    """
    Event types for telegraph events.

    Attributes:
        BUTTON_PRESS: Button make (press) event.
        BUTTON_RELEASE: Button break (release) event.
    """

    BUTTON_PRESS = "M"  # Make
    BUTTON_RELEASE = "B"  # Break



"""
Conbus Event Raw Service for sending raw event telegrams.

This service implements a TCP client that connects to Conbus servers and sends raw event
telegrams to simulate button presses on Conbus modules.
"""

import logging
from typing import Callable, Optional

from twisted.internet.base import DelayedCall

from xp.models import ConbusEventRawResponse
from xp.models.protocol.conbus_protocol import TelegramReceivedEvent
from xp.services.protocol.conbus_event_protocol import ConbusEventProtocol


class ConbusEventRawService:
    """
    Service for sending raw event telegrams to Conbus servers.

    Uses ConbusEventProtocol to send MAKE/BREAK event sequences to
    simulate button presses on Conbus modules.

    Attributes:
        conbus_protocol: Protocol instance for Conbus communication.
    """

    conbus_protocol: ConbusEventProtocol

    def __init__(self, conbus_protocol: ConbusEventProtocol) -> None:
        """
        Initialize the Conbus event raw service.

        Args:
            conbus_protocol: ConbusEventProtocol instance.
        """
        self.progress_callback: Optional[Callable[[str], None]] = None
        self.finish_callback: Optional[Callable[[ConbusEventRawResponse], None]] = None

        self.conbus_protocol: ConbusEventProtocol = conbus_protocol
        self.conbus_protocol.on_connection_made.connect(self.connection_made)
        self.conbus_protocol.on_telegram_sent.connect(self.telegram_sent)
        self.conbus_protocol.on_telegram_received.connect(self.telegram_received)
        self.conbus_protocol.on_timeout.connect(self.timeout)
        self.conbus_protocol.on_failed.connect(self.failed)

        self.event_result = ConbusEventRawResponse(success=False)
        self.logger = logging.getLogger(__name__)

        # Event parameters
        self.module_type_code: int = 0
        self.link_number: int = 0
        self.input_number: int = 0
        self.time_ms: int = 1000
        self.break_event_call: Optional[DelayedCall] = None

    def connection_made(self) -> None:
        """Handle connection established event."""
        self.logger.debug("Connection established")
        self.logger.debug("Sending MAKE event telegram")
        self._send_make_event()

    def _send_make_event(self) -> None:
        """Send MAKE event telegram."""
        payload = f"E{self.module_type_code:02d}L{self.link_number:02d}I{self.input_number:02d}M"
        self.logger.debug(f"Sending MAKE event: {payload}")
        self.conbus_protocol.telegram_queue.put_nowait(payload.encode())
        self.conbus_protocol.call_later(0.0, self.conbus_protocol.start_queue_manager)

        # Schedule BREAK event after delay
        delay_seconds = self.time_ms / 1000.0
        self.break_event_call = self.conbus_protocol.call_later(
            delay_seconds, self._send_break_event
        )

    def _send_break_event(self) -> None:
        """Send BREAK event telegram."""
        payload = f"E{self.module_type_code:02d}L{self.link_number:02d}I{self.input_number:02d}B"
        self.logger.debug(f"Sending BREAK event: {payload}")
        self.conbus_protocol.telegram_queue.put_nowait(payload.encode())
        self.conbus_protocol.call_later(0.0, self.conbus_protocol.start_queue_manager)

    def telegram_sent(self, telegram_sent: str) -> None:
        """
        Handle telegram sent event.

        Args:
            telegram_sent: The telegram that was sent.
        """
        self.logger.debug(f"Telegram sent: {telegram_sent}")
        if self.event_result.sent_telegrams is None:
            self.event_result.sent_telegrams = []
        self.event_result.sent_telegrams.append(telegram_sent)

    def telegram_received(self, telegram_received: TelegramReceivedEvent) -> None:
        """
        Handle telegram received event.

        Args:
            telegram_received: The telegram received event.
        """
        self.logger.debug(f"Telegram received: {telegram_received.frame}")
        if self.event_result.received_telegrams is None:
            self.event_result.received_telegrams = []
        self.event_result.received_telegrams.append(telegram_received.frame)

        # Display progress - show ALL received telegrams
        if self.progress_callback:
            self.progress_callback(telegram_received.frame)

    def timeout(self) -> None:
        """
        Handle timeout event.

        Timeout is the normal/expected way to finish this service.
        """
        timeout_seconds = self.conbus_protocol.timeout_seconds
        self.logger.info("Event raw finished after timeout: %ss", timeout_seconds)
        self.event_result.success = True
        self.event_result.error = None
        if self.finish_callback:
            self.finish_callback(self.event_result)

        self.stop_reactor()

    def failed(self, message: str) -> None:
        """
        Handle failed connection event.

        Args:
            message: Failure message.
        """
        self.logger.debug(f"Failed: {message}")
        self.event_result.success = False
        self.event_result.error = message
        if self.finish_callback:
            self.finish_callback(self.event_result)

        self.stop_reactor()

    def stop_reactor(self) -> None:
        """Stop reactor."""
        self.logger.info("Stopping reactor")
        # Cancel break event call if it's still pending
        if self.break_event_call and self.break_event_call.active():
            self.break_event_call.cancel()
        self.conbus_protocol.stop_reactor()

    def start_reactor(self) -> None:
        """Start reactor."""
        self.logger.info("Starting reactor")
        self.conbus_protocol.start_reactor()

    def run(
        self,
        module_type_code: int,
        link_number: int,
        input_number: int,
        time_ms: int,
        progress_callback: Optional[Callable[[str], None]],
        finish_callback: Callable[[ConbusEventRawResponse], None],
        timeout_seconds: int = 5,
    ) -> None:
        """
        Run reactor in dedicated thread with its own event loop.

        Args:
            module_type_code: Module type code (numeric, e.g., 2 for CP20, 33 for XP33).
            link_number: Link number (0-99).
            input_number: Input number (0-9).
            time_ms: Delay in milliseconds between MAKE and BREAK events.
            progress_callback: Callback for progress updates (received telegrams).
            finish_callback: Callback when operation completes.
            timeout_seconds: Timeout in seconds (default: 5).
        """
        self.logger.info(
            f"Starting event raw: module={module_type_code}, "
            f"link={link_number}, input={input_number}, time={time_ms}ms"
        )

        self.module_type_code = module_type_code
        self.link_number = link_number
        self.input_number = input_number
        self.time_ms = time_ms

        self.conbus_protocol.timeout_seconds = timeout_seconds
        self.progress_callback = progress_callback
        self.finish_callback = finish_callback



"""Unit tests for XP33 Action Table Serializer."""

from contextlib import suppress

import pytest

from xp.models.actiontable.msactiontable_xp33 import (
    Xp33MsActionTable,
    Xp33Output,
    Xp33Scene,
)
from xp.models.telegram.timeparam_type import TimeParam
from xp.services.conbus.actiontable.actiontable_download_service import (
    Xp33MsActionTableSerializer,
)
from xp.utils.serialization import de_nibbles


class TestXp33MsActionTableSerializer:
    """Test cases for Xp33MsActionTableSerializer."""

    @pytest.fixture
    def sample_action_table(self):
        """Create sample action table for testing."""
        return Xp33MsActionTable(
            output1=Xp33Output(
                min_level=10,
                max_level=90,
                scene_outputs=True,
                start_at_full=False,
                leading_edge=True,
            ),
            output2=Xp33Output(
                min_level=20,
                max_level=80,
                scene_outputs=False,
                start_at_full=True,
                leading_edge=False,
            ),
            output3=Xp33Output(
                min_level=30,
                max_level=70,
                scene_outputs=True,
                start_at_full=True,
                leading_edge=True,
            ),
            scene1=Xp33Scene(
                output1_level=50,
                output2_level=60,
                output3_level=70,
                time=TimeParam.T5SEC,
            ),
            scene2=Xp33Scene(
                output1_level=25,
                output2_level=35,
                output3_level=45,
                time=TimeParam.T10SEC,
            ),
            scene3=Xp33Scene(
                output1_level=75,
                output2_level=85,
                output3_level=95,
                time=TimeParam.T1MIN,
            ),
            scene4=Xp33Scene(
                output1_level=0,
                output2_level=100,
                output3_level=50,
                time=TimeParam.NONE,
            ),
        )

    @pytest.fixture
    def sample_telegram_data(self):
        """Sample telegram data for testing - based on specification example."""
        return "AAAABOGEBOGEBOGEAABECIDMAADMFACIAABEBEBEAAGEGEGEAHAAAAAAAAAAAAAAAAAA"

    def test_serialization_round_trip(self, sample_action_table):
        """Test that serialization and deserialization produce the same data."""
        # Serialize to telegram format
        serialized = Xp33MsActionTableSerializer.to_encoded_string(sample_action_table)

        # Deserialize back
        deserialized = Xp33MsActionTableSerializer.from_encoded_string(serialized)

        # Check outputs
        assert deserialized.output1.min_level == sample_action_table.output1.min_level
        assert deserialized.output1.max_level == sample_action_table.output1.max_level
        assert (
            deserialized.output1.scene_outputs
            == sample_action_table.output1.scene_outputs
        )
        assert (
            deserialized.output1.start_at_full
            == sample_action_table.output1.start_at_full
        )
        assert (
            deserialized.output1.leading_edge
            == sample_action_table.output1.leading_edge
        )

        assert deserialized.output2.min_level == sample_action_table.output2.min_level
        assert deserialized.output2.max_level == sample_action_table.output2.max_level
        assert (
            deserialized.output2.scene_outputs
            == sample_action_table.output2.scene_outputs
        )
        assert (
            deserialized.output2.start_at_full
            == sample_action_table.output2.start_at_full
        )
        assert (
            deserialized.output2.leading_edge
            == sample_action_table.output2.leading_edge
        )

        assert deserialized.output3.min_level == sample_action_table.output3.min_level
        assert deserialized.output3.max_level == sample_action_table.output3.max_level
        assert (
            deserialized.output3.scene_outputs
            == sample_action_table.output3.scene_outputs
        )
        assert (
            deserialized.output3.start_at_full
            == sample_action_table.output3.start_at_full
        )
        assert (
            deserialized.output3.leading_edge
            == sample_action_table.output3.leading_edge
        )

        # Check scenes
        assert (
            deserialized.scene1.output1_level
            == sample_action_table.scene1.output1_level
        )
        assert (
            deserialized.scene1.output2_level
            == sample_action_table.scene1.output2_level
        )
        assert (
            deserialized.scene1.output3_level
            == sample_action_table.scene1.output3_level
        )
        assert deserialized.scene1.time == sample_action_table.scene1.time

        assert (
            deserialized.scene2.output1_level
            == sample_action_table.scene2.output1_level
        )
        assert (
            deserialized.scene2.output2_level
            == sample_action_table.scene2.output2_level
        )
        assert (
            deserialized.scene2.output3_level
            == sample_action_table.scene2.output3_level
        )
        assert deserialized.scene2.time == sample_action_table.scene2.time

        assert (
            deserialized.scene3.output1_level
            == sample_action_table.scene3.output1_level
        )
        assert (
            deserialized.scene3.output2_level
            == sample_action_table.scene3.output2_level
        )
        assert (
            deserialized.scene3.output3_level
            == sample_action_table.scene3.output3_level
        )
        assert deserialized.scene3.time == sample_action_table.scene3.time

        assert (
            deserialized.scene4.output1_level
            == sample_action_table.scene4.output1_level
        )
        assert (
            deserialized.scene4.output2_level
            == sample_action_table.scene4.output2_level
        )
        assert (
            deserialized.scene4.output3_level
            == sample_action_table.scene4.output3_level
        )
        assert deserialized.scene4.time == sample_action_table.scene4.time

    def test_from_data_basic(self, sample_telegram_data):
        """Test basic telegram parsing."""
        action_table = Xp33MsActionTableSerializer.from_encoded_string(
            sample_telegram_data
        )

        # Verify it's a valid Xp33MsActionTable
        assert isinstance(action_table, Xp33MsActionTable)

        # Check that we have 3 outputs
        assert action_table.output1 is not None
        assert action_table.output2 is not None
        assert action_table.output3 is not None

        # Check that we have 4 scenes
        assert action_table.scene1 is not None
        assert action_table.scene2 is not None
        assert action_table.scene3 is not None
        assert action_table.scene4 is not None

    def test_from_data_invalid_length(self):
        """Test that invalid telegram length raises ValueError."""
        # Too short
        with pytest.raises(ValueError, match="is too short"):
            Xp33MsActionTableSerializer.from_encoded_string("AAAA")

        with pytest.raises(ValueError, match="is too short"):
            Xp33MsActionTableSerializer.from_encoded_string("AAA")  # Even shorter

    def test_boundary_values(self):
        """Test boundary value handling."""
        # Create action table with boundary values
        action_table = Xp33MsActionTable(
            output1=Xp33Output(min_level=0, max_level=100),
            output2=Xp33Output(min_level=0, max_level=100),
            output3=Xp33Output(min_level=0, max_level=100),
            scene1=Xp33Scene(
                output1_level=0,
                output2_level=100,
                output3_level=50,
                time=TimeParam.NONE,
            ),
            scene2=Xp33Scene(
                output1_level=100,
                output2_level=0,
                output3_level=50,
                time=TimeParam.T120MIN,
            ),
            scene3=Xp33Scene(
                output1_level=50,
                output2_level=50,
                output3_level=50,
                time=TimeParam.T5SEC,
            ),
            scene4=Xp33Scene(
                output1_level=25,
                output2_level=75,
                output3_level=0,
                time=TimeParam.T1MIN,
            ),
        )

        # Test round trip
        serialized = Xp33MsActionTableSerializer.to_encoded_string(action_table)
        deserialized = Xp33MsActionTableSerializer.from_encoded_string(serialized)

        # Verify boundary values are preserved
        assert deserialized.output1.min_level == 0
        assert deserialized.output1.max_level == 100
        assert deserialized.scene1.output1_level == 0
        assert deserialized.scene1.output2_level == 100
        assert deserialized.scene2.output1_level == 100
        assert deserialized.scene2.output2_level == 0

    def test_bit_flags_handling(self):
        """Test proper handling of bit flags for outputs."""
        # Test all combinations of flags
        action_table = Xp33MsActionTable(
            output1=Xp33Output(
                scene_outputs=True, start_at_full=False, leading_edge=True
            ),
            output2=Xp33Output(
                scene_outputs=False, start_at_full=True, leading_edge=False
            ),
            output3=Xp33Output(
                scene_outputs=True, start_at_full=True, leading_edge=True
            ),
        )

        # Test round trip
        serialized = Xp33MsActionTableSerializer.to_encoded_string(action_table)
        deserialized = Xp33MsActionTableSerializer.from_encoded_string(serialized)

        # Verify flags are preserved
        assert deserialized.output1.scene_outputs
        assert not deserialized.output1.start_at_full
        assert deserialized.output1.leading_edge

        assert not deserialized.output2.scene_outputs
        assert deserialized.output2.start_at_full
        assert not deserialized.output2.leading_edge

        assert deserialized.output3.scene_outputs
        assert deserialized.output3.start_at_full
        assert deserialized.output3.leading_edge

    def test_time_param_handling(self):
        """Test TimeParam enum handling."""
        action_table = Xp33MsActionTable(
            scene1=Xp33Scene(time=TimeParam.NONE),
            scene2=Xp33Scene(time=TimeParam.T5SEC),
            scene3=Xp33Scene(time=TimeParam.T1MIN),
            scene4=Xp33Scene(time=TimeParam.T120MIN),
        )

        # Test round trip
        serialized = Xp33MsActionTableSerializer.to_encoded_string(action_table)
        deserialized = Xp33MsActionTableSerializer.from_encoded_string(serialized)

        # Verify time parameters are preserved
        assert deserialized.scene1.time == TimeParam.NONE
        assert deserialized.scene2.time == TimeParam.T5SEC
        assert deserialized.scene3.time == TimeParam.T1MIN
        assert deserialized.scene4.time == TimeParam.T120MIN

    def test_percentage_conversions(self):
        """Test percentage conversion utility methods."""
        # Test percentage to byte
        assert Xp33MsActionTableSerializer._percentage_to_byte(0) == 0
        assert Xp33MsActionTableSerializer._percentage_to_byte(50) == 50
        assert Xp33MsActionTableSerializer._percentage_to_byte(100) == 100
        assert Xp33MsActionTableSerializer._percentage_to_byte(-10) == 0  # Clamped to 0
        assert (
            Xp33MsActionTableSerializer._percentage_to_byte(150) == 100
        )  # Clamped to 100

        # Test byte to percentage
        assert Xp33MsActionTableSerializer._byte_to_percentage(0) == 0
        assert Xp33MsActionTableSerializer._byte_to_percentage(50) == 50
        assert Xp33MsActionTableSerializer._byte_to_percentage(100) == 100
        assert (
            Xp33MsActionTableSerializer._byte_to_percentage(255) == 100
        )  # Clamped to 100

    def test_time_param_conversions(self):
        """Test TimeParam conversion utility methods."""
        # Test time param to byte
        assert Xp33MsActionTableSerializer._time_param_to_byte(TimeParam.NONE) == 0
        assert Xp33MsActionTableSerializer._time_param_to_byte(TimeParam.T5SEC) == 4
        assert Xp33MsActionTableSerializer._time_param_to_byte(TimeParam.T1MIN) == 10

        # Test byte to time param
        assert Xp33MsActionTableSerializer._byte_to_time_param(0) == TimeParam.NONE
        assert Xp33MsActionTableSerializer._byte_to_time_param(4) == TimeParam.T5SEC
        assert Xp33MsActionTableSerializer._byte_to_time_param(10) == TimeParam.T1MIN
        assert (
            Xp33MsActionTableSerializer._byte_to_time_param(255) == TimeParam.NONE
        )  # Invalid value defaults to NONE

    def test_byte_to_bits_conversion(self):
        """Test byte to bits conversion."""
        from xp.utils.serialization import byte_to_bits

        # Test various byte values
        assert byte_to_bits(0) == [False] * 8
        assert byte_to_bits(1) == [True] + [False] * 7
        assert byte_to_bits(2) == [False, True] + [False] * 6
        assert byte_to_bits(255) == [True] * 8

        # Test specific pattern: 0b00000101 = 5
        expected = [True, False, True, False, False, False, False, False]
        assert byte_to_bits(5) == expected

    def test_bits_to_byte_conversion(self):
        """Test bits to byte conversion."""
        from xp.utils.serialization import bits_to_byte

        # Test various bit patterns
        assert bits_to_byte([False] * 8) == 0
        assert bits_to_byte([True] + [False] * 7) == 1
        assert bits_to_byte([False, True] + [False] * 6) == 2
        assert bits_to_byte([True] * 8) == 255

        # Test specific pattern: 0b00000101 = 5
        bits = [True, False, True, False, False, False, False, False]
        assert bits_to_byte(bits) == 5

    def test_exception_handling_for_dim_function(self):
        """Test exception handling for dimFunction bit extraction."""
        # Create a mock raw_bytes that would cause an exception
        # This simulates the exception handling mentioned in the specification
        raw_bytes = bytearray(25)  # Only 25 bytes instead of 32

        # This should not raise an exception and should default to False
        with suppress(IndexError):
            output = Xp33MsActionTableSerializer._decode_output(raw_bytes, 0)
            assert not output.leading_edge

    def test_de_nibble_integration(self):
        """Test integration with de_nibble utility."""
        # Test with simple nibble data
        nibble_data = "AAAB"  # Should decode to [0, 1]
        result = de_nibbles(nibble_data)
        assert result == bytearray([0, 1])

        # Test with longer data
        nibble_data = "AAAAADAAADAAADAAADAAAAAA"
        result = de_nibbles(nibble_data)
        expected = bytearray([0, 0, 3, 0, 3, 0, 3, 0, 3, 0, 0, 0])
        assert result == expected

    def test_default_values(self):
        """Test that default values work correctly."""
        # Create action table with default values
        action_table = Xp33MsActionTable()

        # Test serialization with defaults
        serialized = Xp33MsActionTableSerializer.to_encoded_string(action_table)
        deserialized = Xp33MsActionTableSerializer.from_encoded_string(serialized)

        # Verify defaults are preserved
        assert deserialized.output1.min_level == 0
        assert deserialized.output1.max_level == 100
        assert not deserialized.output1.scene_outputs
        assert not deserialized.output1.start_at_full
        assert not deserialized.output1.leading_edge

        assert deserialized.scene1.output1_level == 0
        assert deserialized.scene1.output2_level == 0
        assert deserialized.scene1.output3_level == 0
        assert deserialized.scene1.time == TimeParam.NONE

    def test_serialize_back_and_forth(self):
        """Test that default values work correctly."""
        telegram = "<R0020045056F17DAAAAAAGEAAGEAAGEAABECIDMAADMFACIAABEBEBEAAGEGEGEAHAAAAAAAAAAAAAAAAAAFI>"
        # Create action table with default values

        # Test serialization with defaults
        serialized_table = telegram[20:84]
        deserialized = Xp33MsActionTableSerializer.from_encoded_string(serialized_table)
        serialized = Xp33MsActionTableSerializer.to_encoded_string(deserialized)

        assert serialized_table == serialized



"""HomeKit configuration models."""

import logging
from pathlib import Path
from typing import List, Optional

from pydantic import BaseModel, IPvAnyAddress


class ConsonModuleConfig(BaseModel):
    """
    Configuration for a Conson module.

    Attributes:
        name: Name of the module.
        serial_number: Serial number of the module.
        module_type: Type of the module.
        module_type_code: Numeric code for the module type.
        link_number: Link number for the module.
        enabled: Whether the module is enabled.
        module_number: Optional module number.
        conbus_ip: Optional Conbus IP address.
        conbus_port: Optional Conbus port number.
        sw_version: Optional software version.
        hw_version: Optional hardware version.
        action_table: Optional action table configuration.
        xp20_msaction_table: Optional xp20 ms action table configuration.
        xp24_msaction_table: Optional xp24 ms action table configuration.
        xp33_msaction_table: Optional xp33 ms action table configuration.
        auto_report_status: Optional auto report status.
    """

    name: str
    serial_number: str
    module_type: str
    module_type_code: int
    link_number: int
    enabled: bool = True
    module_number: Optional[int] = None
    conbus_ip: Optional[IPvAnyAddress] = None
    conbus_port: Optional[int] = None
    sw_version: Optional[str] = None
    hw_version: Optional[str] = None
    auto_report_status: Optional[str] = None
    action_table: Optional[List[str]] = None
    xp20_msaction_table: Optional[List[str]] = None
    xp24_msaction_table: Optional[List[str]] = None
    xp33_msaction_table: Optional[List[str]] = None


class ConsonModuleListConfig(BaseModel):
    """
    Configuration list for Conson modules.

    Attributes:
        root: List of Conson module configurations.
    """

    root: List[ConsonModuleConfig] = []

    @classmethod
    def from_yaml(cls, file_path: str) -> "ConsonModuleListConfig":
        """
        Load configuration from YAML file.

        Args:
            file_path: Path to the YAML configuration file.

        Returns:
            ConsonModuleListConfig instance loaded from file or default config.
        """
        import yaml

        if not Path(file_path).exists():
            logger = logging.getLogger(__name__)
            logger.error(f"File {file_path} does not exist, loading default")
            return cls()

        with Path(file_path).open("r") as file:
            data = yaml.safe_load(file)
        return cls(root=data)

    def find_module(self, serial_number: str) -> Optional[ConsonModuleConfig]:
        """
        Find a module by serial number.

        Args:
            serial_number: Module serial number to search for.

        Returns:
            ConsonModuleConfig if found, None otherwise.
        """
        for module in self.root:
            if module.serial_number == serial_number:
                return module
        return None



"""Unit tests for AccessoryState model."""

from datetime import datetime

import pytest

from xp.models.term.accessory_state import AccessoryState


class TestAccessoryState:
    """Unit tests for AccessoryState dataclass."""

    @pytest.fixture
    def accessory_state(self):
        """Create a basic AccessoryState for testing."""
        return AccessoryState(
            room_name="Living Room",
            accessory_name="Main Light",
            action="a",
            output_state="ON",
            dimming_state="",
            module_name="A01",
            serial_number="1234567890",
            module_type="XP24",
            error_status="OK",
            output=1,
            sort=1,
        )

    def test_initialization(self, accessory_state):
        """Test AccessoryState initializes with correct values."""
        assert accessory_state.room_name == "Living Room"
        assert accessory_state.accessory_name == "Main Light"
        assert accessory_state.action == "a"
        assert accessory_state.output_state == "ON"
        assert accessory_state.dimming_state == ""
        assert accessory_state.module_name == "A01"
        assert accessory_state.serial_number == "1234567890"
        assert accessory_state.module_type == "XP24"
        assert accessory_state.error_status == "OK"
        assert accessory_state.output == 1
        assert accessory_state.sort == 1
        assert accessory_state.last_update is None
        assert accessory_state.toggle_action is None

    def test_initialization_with_optional_fields(self):
        """Test AccessoryState with optional fields set."""
        now = datetime.now()
        state = AccessoryState(
            room_name="Bedroom",
            accessory_name="Dimmer",
            action="b",
            output_state="ON",
            dimming_state="75%",
            module_name="A02",
            serial_number="0987654321",
            module_type="XP33LED",
            error_status="OK",
            output=2,
            sort=2,
            last_update=now,
            toggle_action="E02L01I02",
        )

        assert state.last_update == now
        assert state.toggle_action == "E02L01I02"

    def test_is_dimmable_xp33lr(self):
        """Test is_dimmable returns True for XP33LR modules."""
        state = AccessoryState(
            room_name="Room",
            accessory_name="Dimmer",
            action="a",
            output_state="ON",
            dimming_state="",
            module_name="A01",
            serial_number="1234567890",
            module_type="XP33LR",
            error_status="OK",
            output=1,
            sort=1,
        )
        assert state.is_dimmable() is True

    def test_is_dimmable_xp33led(self):
        """Test is_dimmable returns True for XP33LED modules."""
        state = AccessoryState(
            room_name="Room",
            accessory_name="LED Dimmer",
            action="a",
            output_state="ON",
            dimming_state="",
            module_name="A01",
            serial_number="1234567890",
            module_type="XP33LED",
            error_status="OK",
            output=1,
            sort=1,
        )
        assert state.is_dimmable() is True

    def test_is_dimmable_xp24(self):
        """Test is_dimmable returns False for XP24 modules."""
        state = AccessoryState(
            room_name="Room",
            accessory_name="Switch",
            action="a",
            output_state="ON",
            dimming_state="",
            module_name="A01",
            serial_number="1234567890",
            module_type="XP24",
            error_status="OK",
            output=1,
            sort=1,
        )
        assert state.is_dimmable() is False

    def test_is_dimmable_xp130(self):
        """Test is_dimmable returns False for XP130 modules."""
        state = AccessoryState(
            room_name="Room",
            accessory_name="Sensor",
            action="a",
            output_state="?",
            dimming_state="",
            module_name="A01",
            serial_number="1234567890",
            module_type="XP130",
            error_status="OK",
            output=1,
            sort=1,
        )
        assert state.is_dimmable() is False

    def test_output_states(self):
        """Test various output state values."""
        for output_state in ("ON", "OFF", "?"):
            state = AccessoryState(
                room_name="Room",
                accessory_name="Light",
                action="a",
                output_state=output_state,
                dimming_state="",
                module_name="A01",
                serial_number="1234567890",
                module_type="XP24",
                error_status="OK",
                output=1,
                sort=1,
            )
            assert state.output_state == output_state

    def test_error_status_values(self):
        """Test various error status values."""
        for error_status in ("OK", "E10", "E15"):
            state = AccessoryState(
                room_name="Room",
                accessory_name="Light",
                action="a",
                output_state="ON",
                dimming_state="",
                module_name="A01",
                serial_number="1234567890",
                module_type="XP24",
                error_status=error_status,
                output=1,
                sort=1,
            )
            assert state.error_status == error_status



"""Log entry model for console bus communication logs."""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Optional, Union

from xp.models.telegram.event_telegram import EventTelegram
from xp.models.telegram.reply_telegram import ReplyTelegram
from xp.models.telegram.system_telegram import SystemTelegram


@dataclass
class LogEntry:
    """
    Represents a single entry in a console bus log file.

    Format: HH:MM:SS,mmm [TX/RX] <telegram>
    Examples: 22:44:20,352 [TX] <S0012345008F27D00AAFN>

    Attributes:
        timestamp: Timestamp of the log entry.
        direction: Direction of telegram ("TX" or "RX").
        raw_telegram: Raw telegram string.
        parsed_telegram: Parsed telegram object if successfully parsed.
        parse_error: Error message if parsing failed.
        line_number: Line number in the log file.
        is_transmitted: True if this is a transmitted telegram.
        is_received: True if this is a received telegram.
        telegram_type: Telegram type (event, system, reply, unknown).
        is_valid_parse: True if the telegram was successfully parsed.
        checksum_validated: Checksum validation status if available.
    """

    timestamp: datetime
    direction: str  # "TX" or "RX"
    raw_telegram: str
    parsed_telegram: Optional[Union[EventTelegram, SystemTelegram, ReplyTelegram]] = (
        None
    )
    parse_error: Optional[str] = None
    line_number: int = 0

    @property
    def is_transmitted(self) -> bool:
        """
        True if this is a transmitted telegram.

        Returns:
            True if direction is TX, False otherwise.
        """
        return self.direction == "TX"

    @property
    def is_received(self) -> bool:
        """
        True if this is a received telegram.

        Returns:
            True if direction is RX, False otherwise.
        """
        return self.direction == "RX"

    @property
    def telegram_type(self) -> str:
        """
        Get the telegram type (event, system, reply, unknown).

        Returns:
            Telegram type string.
        """
        if self.parsed_telegram is None:
            return "unknown"

        return self.parsed_telegram.telegram_type.value.lower()

    @property
    def is_valid_parse(self) -> bool:
        """
        True if the telegram was successfully parsed.

        Returns:
            True if parsed without errors, False otherwise.
        """
        return self.parsed_telegram is not None and self.parse_error is None

    @property
    def checksum_validated(self) -> Optional[bool]:
        """
        Get checksum validation status if available.

        Returns:
            Checksum validation status or None if not available.
        """
        if self.parsed_telegram and hasattr(self.parsed_telegram, "checksum_validated"):
            return self.parsed_telegram.checksum_validated
        return None

    def to_dict(self) -> dict[str, Any]:
        """
        Convert to dictionary for JSON serialization.

        Returns:
            Dictionary representation of the log entry.
        """
        result: dict[str, Any] = {
            "line_number": self.line_number,
            "timestamp": self.timestamp.strftime("%H:%M:%S.%f")[
                :-3
            ],  # HH:MM:SS,mmm format
            "direction": self.direction,
            "raw_telegram": self.raw_telegram,
            "telegram_type": self.telegram_type,
            "is_valid_parse": self.is_valid_parse,
            "parse_error": self.parse_error,
        }

        # Add parsed telegram data if available
        if self.parsed_telegram:
            result["parsed"] = self.parsed_telegram.to_dict()
            result["checksum_validated"] = self.checksum_validated

        return result

    def __str__(self) -> str:
        """
        Human-readable string representation.

        Returns:
            Formatted string representation of the log entry.
        """
        timestamp_str = self.timestamp.strftime("%H:%M:%S,%f")[:-3]  # HH:MM:SS,mmm
        status = "✓" if self.is_valid_parse else "✗"
        checksum_status = ""

        if self.checksum_validated is not None:
            checksum_status = f" ({('✓' if self.checksum_validated else '✗')})"

        return f"[{self.line_number:3d}] {timestamp_str} [{self.direction}] {self.raw_telegram} {status}{checksum_status}"



"""
XP20 Server Service for device emulation.

This service provides XP20-specific device emulation functionality, including response
generation and device configuration handling.
"""

from typing import Dict, Optional

from xp.models import ModuleTypeCode
from xp.models.actiontable.msactiontable_xp20 import Xp20MsActionTable
from xp.services.actiontable.msactiontable_xp20_serializer import (
    Xp20MsActionTableSerializer,
)
from xp.services.server.base_server_service import BaseServerService


class XP20ServerError(Exception):
    """Raised when XP20 server operations fail."""

    pass


class XP20ServerService(BaseServerService):
    """
    XP20 device emulation service.

    Generates XP20-specific responses, handles XP20 device configuration, and implements
    XP20 telegram format.
    """

    def __init__(
        self,
        serial_number: str,
        _variant: str = "",
        msactiontable_serializer: Optional[Xp20MsActionTableSerializer] = None,
    ):
        """
        Initialize XP20 server service.

        Args:
            serial_number: The device serial number.
            _variant: Reserved parameter for consistency (unused).
            msactiontable_serializer: MsActionTable serializer (injected via DI).
        """
        super().__init__(serial_number)
        self.device_type = "XP20"
        self.module_type_code = ModuleTypeCode.XP20  # XP20 module type from registry
        self.firmware_version = "XP20_V0.01.05"

        # MsActionTable support
        self.msactiontable_serializer = (
            msactiontable_serializer or Xp20MsActionTableSerializer()
        )
        self.msactiontable = self._get_default_msactiontable()

    def _get_msactiontable_serializer(self) -> Optional[Xp20MsActionTableSerializer]:
        """
        Get the MsActionTable serializer for XP20.

        Returns:
            The XP20 MsActionTable serializer instance.
        """
        return self.msactiontable_serializer

    def _get_msactiontable(self) -> Optional[Xp20MsActionTable]:
        """
        Get the MsActionTable for XP20.

        Returns:
            The XP20 MsActionTable instance.
        """
        return self.msactiontable

    def _get_default_msactiontable(self) -> Xp20MsActionTable:
        """
        Generate default MsActionTable configuration.

        Returns:
            Default XP20 MsActionTable with all inputs unconfigured.
        """
        # All inputs unconfigured (all flags False, AND functions empty)
        return Xp20MsActionTable()

    def get_device_info(self) -> Dict:
        """
        Get XP20 device information.

        Returns:
            Dictionary containing device information.
        """
        return {
            "serial_number": self.serial_number,
            "device_type": self.device_type,
            "firmware_version": self.firmware_version,
            "status": self.device_status,
            "link_number": self.link_number,
        }



"""Unit tests for ConbusEventProtocol."""

import asyncio
from unittest.mock import Mock, patch

import pytest

from xp.models import ConbusClientConfig
from xp.services.protocol.conbus_event_protocol import ConbusEventProtocol


class TestConbusEventProtocol:
    """Unit tests for ConbusEventProtocol functionality."""

    @pytest.fixture
    def mock_reactor(self):
        """Create a mock reactor."""
        reactor = Mock()
        reactor.connectTCP = Mock()
        reactor.running = False
        return reactor

    @pytest.fixture
    def mock_cli_config(self):
        """Create a mock CLI config."""
        config = Mock(spec=ConbusClientConfig)
        config.conbus = Mock()
        config.conbus.ip = "192.168.1.100"
        config.conbus.port = 10001
        config.conbus.timeout = 5.0
        return config

    @pytest.fixture
    def mock_telegram_service(self):
        """Create a mock telegram service."""
        return Mock()

    @pytest.fixture
    def protocol(self, mock_cli_config, mock_reactor, mock_telegram_service):
        """Create protocol instance with mocks."""
        return ConbusEventProtocol(
            cli_config=mock_cli_config,
            reactor=mock_reactor,
            telegram_service=mock_telegram_service,
        )

    def test_protocol_initialization(self, protocol, mock_cli_config, mock_reactor):
        """Test protocol can be initialized with required dependencies."""
        assert protocol.buffer == b""
        assert protocol._reactor == mock_reactor
        assert protocol.cli_config == mock_cli_config.conbus
        assert protocol.timeout_seconds == 5.0

    def test_connect_without_running_event_loop(self, protocol, mock_reactor):
        """Test connect() works without running asyncio event loop."""
        # Call connect - should not raise exception
        protocol.connect()

        # Verify connectTCP was called with correct parameters
        mock_reactor.connectTCP.assert_called_once_with(
            "192.168.1.100", 10001, protocol
        )

    @pytest.mark.anyio(backends=["asyncio"])
    async def test_connect_with_running_event_loop(self, protocol, mock_reactor):
        """Test connect() auto-detects and sets running asyncio event loop."""
        with patch.object(protocol, "set_event_loop") as mock_set_event_loop:
            # Call connect within async context (event loop is running)
            protocol.connect()

            # Verify set_event_loop was called with the running event loop
            mock_set_event_loop.assert_called_once()
            call_args = mock_set_event_loop.call_args[0][0]
            assert isinstance(call_args, asyncio.AbstractEventLoop)

            # Verify connectTCP was still called
            mock_reactor.connectTCP.assert_called_once_with(
                "192.168.1.100", 10001, protocol
            )

    def test_set_event_loop(self, protocol, mock_reactor):
        """Test set_event_loop properly configures reactor."""
        # Create a real event loop
        loop = asyncio.new_event_loop()

        try:
            # Mock reactor attributes
            mock_reactor._asyncioEventloop = None
            mock_reactor.running = False
            mock_reactor.startRunning = Mock()

            protocol.set_event_loop(loop)

            # Verify reactor was configured
            assert mock_reactor._asyncioEventloop == loop
            assert mock_reactor.running is True
            mock_reactor.startRunning.assert_called_once()
        finally:
            loop.close()

    def test_set_event_loop_without_startRunning_method(self, protocol, mock_reactor):
        """Test set_event_loop works even if reactor lacks startRunning method."""
        # Create a real event loop
        loop = asyncio.new_event_loop()

        try:
            # Mock reactor attributes (no startRunning method)
            mock_reactor._asyncioEventloop = None
            mock_reactor.running = False
            # Don't add startRunning method

            protocol.set_event_loop(loop)

            # Verify reactor was configured
            assert mock_reactor._asyncioEventloop == loop
            assert mock_reactor.running is True
        finally:
            loop.close()

    def test_disconnect(self, protocol, mock_reactor):
        """Test disconnect calls reactor.disconnectAll."""
        protocol.disconnect()

        mock_reactor.disconnectAll.assert_called_once()

    def test_stop_reactor_when_running(self, protocol, mock_reactor):
        """Test stop_reactor stops the reactor if it's running."""
        mock_reactor.running = True
        mock_reactor.stop = Mock()

        protocol.stop_reactor()

        mock_reactor.stop.assert_called_once()

    def test_stop_reactor_when_not_running(self, protocol, mock_reactor):
        """Test stop_reactor does nothing if reactor is not running."""
        mock_reactor.running = False
        mock_reactor.stop = Mock()

        protocol.stop_reactor()

        mock_reactor.stop.assert_not_called()

    def test_send_raw_telegram(self, protocol):
        """Test send_raw_telegram queues telegram."""
        with patch.object(protocol, "call_later") as mock_call_later:
            protocol.send_raw_telegram("E02L01I08M")

            # Verify telegram was queued
            assert not protocol.telegram_queue.empty()
            telegram = protocol.telegram_queue.get_nowait()
            assert telegram == b"E02L01I08M"

            # Verify queue manager was scheduled
            mock_call_later.assert_called_once()



"""Unit tests for TelegramOutputService.format_output_state method."""

from xp.services.telegram.telegram_output_service import TelegramOutputService


class TestFormatOutputState:
    """Test cases for format_output_state static method."""

    def test_format_xxxx0101(self):
        """Test format_output_state with 'xxxx0101'."""
        result = TelegramOutputService.format_output_state("xxxx0101")
        assert result == "1 0 1 0"

    def test_format_xxxx1110(self):
        """Test format_output_state with 'xxxx1110'."""
        result = TelegramOutputService.format_output_state("xxxx1110")
        assert result == "0 1 1 1"

    def test_format_xxxx0001(self):
        """Test format_output_state with 'xxxx0001'."""
        result = TelegramOutputService.format_output_state("xxxx0001")
        assert result == "1 0 0 0"

    def test_format_xxxx01(self):
        """Test format_output_state with 'xxxx01' (short input)."""
        result = TelegramOutputService.format_output_state("xxxx01")
        # "xxxx01" -> remove x -> "01" -> pad right -> "01  " -> invert -> "  10" -> spaces -> "    1 0"
        assert result == "    1 0"

    def test_format_xx1110(self):
        """Test format_output_state with 'xx1110'."""
        result = TelegramOutputService.format_output_state("xx1110")
        assert result == "0 1 1 1"

    def test_format_0000(self):
        """Test format_output_state with '0000' (no x)."""
        result = TelegramOutputService.format_output_state("0000")
        assert result == "0 0 0 0"

    def test_format_1111(self):
        """Test format_output_state with '1111' (no x)."""
        result = TelegramOutputService.format_output_state("1111")
        assert result == "1 1 1 1"

    def test_format_uppercase_X(self):
        """Test format_output_state with uppercase 'X'."""
        result = TelegramOutputService.format_output_state("XXXX0101")
        assert result == "1 0 1 0"

    def test_format_mixed_case(self):
        """Test format_output_state with mixed case 'xX'."""
        result = TelegramOutputService.format_output_state("xXxX1010")
        assert result == "0 1 0 1"

    def test_format_empty_string(self):
        """Test format_output_state with empty string."""
        result = TelegramOutputService.format_output_state("")
        assert result == "       "  # 4 spaces joined = 7 chars with spaces

    def test_format_single_digit(self):
        """Test format_output_state with single digit."""
        result = TelegramOutputService.format_output_state("1")
        # "1" -> pad right -> "1   " -> invert -> "   1" -> spaces -> "      1"
        assert result == "      1"

    def test_format_xxxx0053(self):
        """Test format_output_state with 'xxxx0053' - real world example."""
        result = TelegramOutputService.format_output_state("xxxx0053")
        assert result == "3 5 0 0"

    def test_format_real_telegram_1235xxxxx000(self):
        """
        Test format_output_state with real telegram data '1235xxxxx000'.

        From telegram: <R0020045057F02D1235xxxxx000BO>
        Data value: 1235xxxxx000
        Algorithm:
        1. Remove 'x': "1235000"
        2. Pad/truncate to 4 chars: "1235" (first 4 chars)
        3. Invert: "5321"
        4. Add spaces: "5 3 2 1"
        """
        result = TelegramOutputService.format_output_state("1235xxxxx000")
        assert result == "5 3 2 1"
